"""Address parsing and matching for census-lookup."""

from census_lookup.address.matcher import GeocodingResult, TIGERAddressMatcher
from census_lookup.address.normalizer import StreetNormalizer
from census_lookup.address.parser import AddressParser, ParsedAddress

__all__ = [
    "AddressParser",
    "ParsedAddress",
    "StreetNormalizer",
    "TIGERAddressMatcher",
    "GeocodingResult",
]



"""CLI commands for census-lookup."""

import asyncio
import json
from pathlib import Path
from typing import Optional

import click
import pandas as pd

from census_lookup import CensusLookup, GeoLevel
from census_lookup.census.variables import VARIABLES, list_variable_groups
from census_lookup.data.constants import FIPS_STATES, normalize_state
from census_lookup.data.manager import DataManager


@click.group()
@click.version_option(package_name="census-lookup")
def cli():
    """Census-lookup: Map US addresses to Census 2020 data."""
    pass


@cli.command()
@click.argument("address")
@click.option(
    "--level",
    "-l",
    default="block",
    type=click.Choice(["block", "block_group", "tract", "county", "state"]),
    help="Geographic level for results",
)
@click.option(
    "--variables",
    "-v",
    multiple=True,
    help="Census variables to include (can be specified multiple times)",
)
@click.option(
    "--format",
    "-f",
    "output_format",
    default="json",
    type=click.Choice(["json", "csv", "table"]),
    help="Output format",
)
@click.option(
    "--no-download",
    is_flag=True,
    help="Don't auto-download missing data",
)
def lookup(address: str, level: str, variables: tuple, output_format: str, no_download: bool):
    """Look up census data for a single address."""
    asyncio.run(_lookup_async(address, level, variables, output_format, no_download))


async def _lookup_async(
    address: str, level: str, variables: tuple, output_format: str, no_download: bool
):
    """Async implementation of lookup command."""
    geo_level = GeoLevel[level.upper()]

    lookup_instance = CensusLookup(
        geo_level=geo_level,
        variables=list(variables) if variables else None,
        auto_download=not no_download,
    )

    try:
        result = await lookup_instance.geocode(address)

        if output_format == "json":
            click.echo(json.dumps(result.to_dict(), indent=2, default=str))
        elif output_format == "csv":
            click.echo(result.to_series().to_csv())
        else:
            _print_result_table(result)
    finally:
        await lookup_instance.close()


def _print_result_table(result):
    """Print a result as a formatted table."""
    data = result.to_dict()

    click.echo("\n" + "=" * 50)
    click.echo("CENSUS LOOKUP RESULT")
    click.echo("=" * 50)

    if result.is_matched:
        click.echo(f"\nInput:    {data['input_address']}")
        click.echo(f"Matched:  {data['matched_address']}")
        click.echo(f"Location: ({data['latitude']:.6f}, {data['longitude']:.6f})")
        click.echo(f"GEOID:    {data['geoid']}")
        click.echo(f"Match:    {data['match_type']} (score: {data['match_score']:.2f})")

        if result.census_data:
            click.echo("\nCensus Data:")
            for var, value in result.census_data.items():
                desc = VARIABLES.get(var, var)
                click.echo(f"  {var}: {value} ({desc})")
    else:
        click.echo(f"\nInput: {data['input_address']}")
        click.echo(f"Status: {data['match_type']}")
        click.echo("No match found.")

    click.echo("")


@cli.command()
@click.argument("input_file", type=click.Path(exists=True))
@click.argument("output_file", type=click.Path())
@click.option(
    "--address-column",
    "-a",
    required=True,
    help="Column containing addresses",
)
@click.option(
    "--level",
    "-l",
    default="block",
    type=click.Choice(["block", "block_group", "tract", "county"]),
)
@click.option("--variables", "-v", multiple=True, help="Census variables to include")
@click.option("--no-download", is_flag=True, help="Don't auto-download missing data")
def batch(
    input_file: str,
    output_file: str,
    address_column: str,
    level: str,
    variables: tuple,
    no_download: bool,
):
    """Process a batch of addresses from CSV file."""
    asyncio.run(
        _batch_async(input_file, output_file, address_column, level, variables, no_download)
    )


async def _batch_async(
    input_file: str,
    output_file: str,
    address_column: str,
    level: str,
    variables: tuple,
    no_download: bool,
):
    """Async implementation of batch command."""
    input_path = Path(input_file)

    # Read input
    if input_path.suffix != ".csv":
        raise click.ClickException(
            f"Unsupported file format: {input_path.suffix}. Only CSV is supported."
        )
    df = pd.read_csv(input_path)

    if address_column not in df.columns:
        raise click.ClickException(f"Column '{address_column}' not found in input file")

    # Process
    geo_level = GeoLevel[level.upper()]
    lookup_instance = CensusLookup(
        geo_level=geo_level,
        variables=list(variables) if variables else None,
        auto_download=not no_download,
    )

    try:
        results = await lookup_instance.geocode_batch(
            df[address_column],
            progress=True,
        )

        # Merge results with original data
        output_df = pd.concat(
            [df.reset_index(drop=True), results.reset_index(drop=True)], axis=1
        )

        # Save
        output_path = Path(output_file)
        if output_path.suffix == ".csv":
            output_df.to_csv(output_path, index=False)
        elif output_path.suffix == ".parquet":
            output_df.to_parquet(output_path)
        else:
            output_df.to_csv(output_path, index=False)

        click.echo(f"Processed {len(df)} addresses -> {output_file}")

        # Summary
        matched = results["match_type"].isin(["interpolated", "exact"]).sum()
        click.echo(f"Matched: {matched}/{len(df)} ({100*matched/len(df):.1f}%)")
    finally:
        await lookup_instance.close()


@cli.command()
@click.argument("states", nargs=-1)
@click.option("--all", "download_all", is_flag=True, help="Download data for all states")
def download(states: tuple, download_all: bool):
    """Pre-download data for specified states."""
    asyncio.run(_download_async(states, download_all))


async def _download_async(states: tuple, download_all: bool):
    """Async implementation of download command."""
    manager = DataManager()

    if download_all:
        states = tuple(FIPS_STATES.keys())

    if not states:
        raise click.ClickException("Specify state(s) to download or use --all")

    try:
        for state in states:
            try:
                state_fips = normalize_state(state)
                state_name = FIPS_STATES.get(state_fips, state)
                click.echo(f"\nDownloading data for {state_name} ({state_fips})...")
                await manager.ensure_state_data(state_fips, show_progress=True)
                click.echo("  Done.")
            except Exception as e:
                click.echo(f"  Error: {e}", err=True)

        click.echo("\nDownload complete!")
    finally:
        await manager.close()


@cli.command()
def info():
    """Show information about downloaded data."""
    manager = DataManager()

    click.echo(f"\nData directory: {manager.data_dir}")

    # Disk usage
    usage = manager.disk_usage()
    click.echo("\nDisk usage:")
    for category, size in usage.items():
        if category != "total":
            click.echo(f"  {category}: {_format_size(size)}")
    click.echo(f"  Total: {_format_size(usage['total'])}")

    # Available states
    states = manager.list_available_states("blocks")
    if states:
        click.echo(f"\nStates with block data ({len(states)}):")
        state_names = [FIPS_STATES.get(s, s) for s in states]
        click.echo(f"  {', '.join(state_names)}")
    else:
        click.echo("\nNo data downloaded yet.")
        click.echo("Run: census-lookup download <state>")


def _format_size(size_bytes: int) -> str:
    """Format bytes as human-readable string."""
    for unit in ["B", "KB", "MB", "GB"]:
        if size_bytes < 1024:
            return f"{size_bytes:.1f} {unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f} TB"


@cli.command()
@click.option("--table", "-t", help="Filter by table (P1, P2, P3, P4, H1)")
def variables(table: Optional[str]):
    """List available census variables."""
    click.echo("\nAvailable Census Variables (PL 94-171):")
    click.echo("=" * 60)

    for code, desc in sorted(VARIABLES.items()):
        if table and not code.startswith(table):
            continue
        click.echo(f"{code:12s} {desc}")

    click.echo("\n\nVariable Groups:")
    click.echo("-" * 40)
    for group, desc in list_variable_groups().items():
        click.echo(f"{group:20s} {desc}")


@cli.command()
@click.argument("state", required=False)
def clear(state: Optional[str]):
    """Clear cached data for a state or all states."""
    manager = DataManager()

    if state:
        state_fips = normalize_state(state)
        click.confirm(
            f"Clear all cached data for {FIPS_STATES.get(state_fips, state)}?",
            abort=True,
        )
        manager.clear_cache(state_fips)
        click.echo("Cache cleared.")
    else:
        click.confirm("Clear ALL cached data?", abort=True)
        manager.clear_cache()
        click.echo("All cache cleared.")


@cli.command()
@click.argument("lat", type=float)
@click.argument("lon", type=float)
@click.option(
    "--level",
    "-l",
    default="block",
    type=click.Choice(["block", "block_group", "tract", "county"]),
)
@click.option("--variables", "-v", multiple=True, help="Census variables to include")
@click.option("--no-download", is_flag=True, help="Don't auto-download missing data")
def coords(lat: float, lon: float, level: str, variables: tuple, no_download: bool):
    """Look up census data for coordinates (lat, lon)."""
    asyncio.run(_coords_async(lat, lon, level, variables, no_download))


async def _coords_async(
    lat: float, lon: float, level: str, variables: tuple, no_download: bool
):
    """Async implementation of coords command."""
    geo_level = GeoLevel[level.upper()]

    lookup_instance = CensusLookup(
        geo_level=geo_level,
        variables=list(variables) if variables else None,
        auto_download=not no_download,
    )

    try:
        # Need to load states that might contain these coordinates
        click.echo("Attempting to find containing state...")

        result = await lookup_instance.lookup_coordinates(lat, lon)

        if result.is_matched:
            click.echo(json.dumps(result.to_dict(), indent=2, default=str))
        else:
            click.echo("No census block found for these coordinates.")
            click.echo("Try loading the appropriate state first: census-lookup download <state>")
    finally:
        await lookup_instance.close()


if __name__ == "__main__":
    cli()



"""Data catalog for tracking downloaded datasets."""

import json
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional


@dataclass
class DatasetInfo:
    """Metadata about a downloaded dataset."""

    dataset_type: str  # "blocks", "addrfeat", "pl94171", etc.
    state_fips: str
    county_fips: Optional[str]  # None for state-level datasets
    file_path: str
    downloaded_at: str  # ISO format datetime
    file_size: int  # bytes
    source_url: str
    checksum: Optional[str] = None

    @classmethod
    def create(
        cls,
        dataset_type: str,
        state_fips: str,
        file_path: Path,
        source_url: str,
        county_fips: Optional[str] = None,
    ) -> "DatasetInfo":
        """Create a new DatasetInfo with current timestamp."""
        return cls(
            dataset_type=dataset_type,
            state_fips=state_fips,
            county_fips=county_fips,
            file_path=str(file_path),
            downloaded_at=datetime.now().isoformat(),
            file_size=file_path.stat().st_size if file_path.exists() else 0,
            source_url=source_url,
        )


@dataclass
class CatalogData:
    """Root catalog data structure."""

    version: str = "1.0"
    datasets: Dict[str, DatasetInfo] = field(default_factory=dict)


class DataCatalog:
    """
    Tracks downloaded and converted data.

    Persists to catalog.json for cross-session awareness.
    """

    def __init__(self, catalog_path: Path):
        """
        Initialize catalog.

        Args:
            catalog_path: Path to catalog.json file
        """
        self.catalog_path = catalog_path
        self._data = CatalogData()
        self._load()

    def _load(self) -> None:
        """Load catalog from disk."""
        if self.catalog_path.exists():
            try:
                with open(self.catalog_path, "r") as f:
                    raw = json.load(f)
                    self._data.version = raw.get("version", "1.0")
                    datasets = raw.get("datasets", {})
                    self._data.datasets = {
                        k: DatasetInfo(**v) for k, v in datasets.items()
                    }
            except (json.JSONDecodeError, TypeError):
                # Corrupted catalog, start fresh
                self._data = CatalogData()

    def _save(self) -> None:
        """Persist catalog to disk."""
        self.catalog_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.catalog_path, "w") as f:
            data = {
                "version": self._data.version,
                "datasets": {k: asdict(v) for k, v in self._data.datasets.items()},
            }
            json.dump(data, f, indent=2)

    def _make_key(
        self,
        dataset_type: str,
        state_fips: str,
        county_fips: Optional[str] = None,
    ) -> str:
        """Create a unique key for a dataset."""
        if county_fips:
            return f"{dataset_type}:{state_fips}:{county_fips}"
        return f"{dataset_type}:{state_fips}"

    def register(self, info: DatasetInfo) -> None:
        """
        Register a downloaded dataset.

        Args:
            info: Dataset metadata
        """
        key = self._make_key(info.dataset_type, info.state_fips, info.county_fips)
        self._data.datasets[key] = info
        self._save()

    def unregister(
        self,
        dataset_type: str,
        state_fips: str,
        county_fips: Optional[str] = None,
    ) -> None:
        """Remove a dataset from the catalog."""
        key = self._make_key(dataset_type, state_fips, county_fips)
        if key in self._data.datasets:
            del self._data.datasets[key]
            self._save()

    def is_available(
        self,
        dataset_type: str,
        state_fips: str,
        county_fips: Optional[str] = None,
    ) -> bool:
        """
        Check if a dataset is available locally.

        Args:
            dataset_type: Type of dataset
            state_fips: State FIPS code
            county_fips: Optional county FIPS code

        Returns:
            True if dataset is registered and file exists
        """
        key = self._make_key(dataset_type, state_fips, county_fips)
        if key not in self._data.datasets:
            return False

        # Verify file still exists
        info = self._data.datasets[key]
        path = Path(info.file_path)
        return path.exists()

    def get_info(
        self,
        dataset_type: str,
        state_fips: str,
        county_fips: Optional[str] = None,
    ) -> Optional[DatasetInfo]:
        """Get metadata for a dataset if available."""
        key = self._make_key(dataset_type, state_fips, county_fips)
        return self._data.datasets.get(key)

    def get_path(
        self,
        dataset_type: str,
        state_fips: str,
        county_fips: Optional[str] = None,
    ) -> Optional[Path]:
        """
        Get path to a dataset if available.

        Returns:
            Path to dataset file, or None if not available
        """
        info = self.get_info(dataset_type, state_fips, county_fips)
        if info:
            path = Path(info.file_path)
            if path.exists():
                return path
        return None

    def list_states(self, dataset_type: str) -> List[str]:
        """
        List states with available data for a dataset type.

        Args:
            dataset_type: Type of dataset

        Returns:
            List of state FIPS codes
        """
        states = set()
        for key, info in self._data.datasets.items():
            if info.dataset_type == dataset_type:
                states.add(info.state_fips)
        return sorted(states)

    def clear(self, dataset_type: Optional[str] = None, state_fips: Optional[str] = None) -> None:
        """
        Clear catalog entries.

        Args:
            dataset_type: Filter by type (None for all)
            state_fips: Filter by state (None for all)
        """
        keys_to_remove = []
        for key, info in self._data.datasets.items():
            if dataset_type and info.dataset_type != dataset_type:
                continue
            if state_fips and info.state_fips != state_fips:
                continue
            keys_to_remove.append(key)

        for key in keys_to_remove:
            del self._data.datasets[key]

        self._save()



"""Match addresses to TIGER address range segments."""

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import geopandas as gpd
from shapely.geometry import LineString, Point

from census_lookup.address.normalizer import StreetNormalizer
from census_lookup.address.parser import AddressParser, ParsedAddress


@dataclass
class GeocodingResult:
    """Result from address geocoding."""

    latitude: Optional[float] = None
    longitude: Optional[float] = None
    matched_address: Optional[str] = None
    match_type: str = "no_match"  # "exact", "interpolated", "no_match"
    match_score: float = 0.0  # 0.0 to 1.0
    tiger_line_id: Optional[str] = None
    side: Optional[str] = None  # "L" or "R"

    @property
    def is_matched(self) -> bool:
        """Check if geocoding was successful."""
        return self.match_type != "no_match" and self.latitude is not None


class TIGERAddressMatcher:
    """
    Match addresses to TIGER address range segments.

    Uses address interpolation along street segments based on
    address ranges encoded in TIGER/Line ADDRFEAT files.

    TIGER ADDRFEAT files contain:
    - FULLNAME: Full street name
    - LFROMHN, LTOHN: Left side house number range
    - RFROMHN, RTOHN: Right side house number range
    - ZIPL, ZIPR: ZIP codes for each side
    - PARITYL, PARITYR: Address parity (O=odd, E=even, B=both)
    - geometry: LineString of street segment
    """

    def __init__(self, addr_features: gpd.GeoDataFrame):
        """
        Initialize with TIGER address features.

        Args:
            addr_features: GeoDataFrame from ADDRFEAT files
        """
        self._features = addr_features.copy()
        self._normalizer = StreetNormalizer()
        self._parser = AddressParser()
        self._build_indexes()

    def _build_indexes(self) -> None:
        """Build indexes for fast street name lookup."""
        # Create normalized street name column
        self._features["_norm_name"] = self._features["FULLNAME"].apply(
            lambda x: self._normalizer.normalize(str(x) if x else "", expand_abbreviations=False)
        )

        # Build dict: normalized_name -> list of row indices
        self._street_index: Dict[str, List[int]] = {}
        for idx, name in enumerate(self._features["_norm_name"]):
            if not name:
                continue
            if name not in self._street_index:
                self._street_index[name] = []
            self._street_index[name].append(idx)

        # Build ZIP code index for faster filtering
        self._zip_index: Dict[str, List[int]] = {}
        for idx, row in self._features.iterrows():
            for zip_col in ["ZIPL", "ZIPR"]:
                zipcode = row.get(zip_col)
                if zipcode:
                    zipcode = str(zipcode).strip()
                    if zipcode not in self._zip_index:
                        self._zip_index[zipcode] = []
                    self._zip_index[zipcode].append(idx)

    def geocode_parsed(self, parsed: ParsedAddress) -> GeocodingResult:
        """
        Geocode a parsed address.

        Args:
            parsed: Parsed address components

        Returns:
            GeocodingResult with coordinates and match info
        """
        if not parsed.has_street_info:
            return GeocodingResult(match_type="no_match", match_score=0.0)

        try:
            house_number = int(parsed.house_number)
        except (ValueError, TypeError):
            return GeocodingResult(match_type="no_match", match_score=0.0)

        # Build normalized street name for matching
        # TIGER uses abbreviated format, so normalize without expansion
        street_name = self._normalizer.normalize(
            parsed.full_street_name, expand_abbreviations=False
        )

        # Find matching segment
        result = self._find_segment(
            house_number=house_number,
            street_name=street_name,
            zipcode=parsed.zipcode,
        )

        if result is None:
            # Try variants
            for variant in self._normalizer.generate_variants(street_name):
                result = self._find_segment(
                    house_number=house_number,
                    street_name=variant,
                    zipcode=parsed.zipcode,
                )
                if result:
                    break

        if result is None:
            return GeocodingResult(match_type="no_match", match_score=0.0)

        segment, side, from_addr, to_addr, tiger_id = result

        # Interpolate position
        point = self._interpolate_position(
            segment_geom=segment.geometry,
            house_number=house_number,
            from_addr=from_addr,
            to_addr=to_addr,
            side=side,
        )

        return GeocodingResult(
            latitude=point.y,
            longitude=point.x,
            matched_address=segment.get("FULLNAME"),
            match_type="interpolated",
            match_score=0.9,  # Could be refined based on match quality
            tiger_line_id=tiger_id,
            side=side,
        )

    def _find_segment(
        self,
        house_number: int,
        street_name: str,
        zipcode: Optional[str] = None,
    ) -> Optional[Tuple[gpd.GeoSeries, str, int, int, str]]:
        """
        Find the street segment containing an address.

        Args:
            house_number: House number to find
            street_name: Normalized street name
            zipcode: Optional ZIP code for filtering

        Returns:
            Tuple of (segment, side, from_addr, to_addr, tiger_id) or None
        """
        # Get candidate segments by street name
        candidates_idx = self._street_index.get(street_name, [])

        if not candidates_idx:
            return None

        candidates = self._features.iloc[candidates_idx]

        # Filter by ZIP if provided
        if zipcode:
            zipcode = str(zipcode).strip()
            zip_mask = (candidates["ZIPL"].astype(str) == zipcode) | (
                candidates["ZIPR"].astype(str) == zipcode
            )
            if zip_mask.any():
                candidates = candidates[zip_mask]

        # Find segment with matching address range
        for idx, segment in candidates.iterrows():
            match = self._check_range(segment, house_number)
            if match:
                side, from_addr, to_addr = match
                tiger_id = segment.get("LINEARID", str(idx))
                return segment, side, from_addr, to_addr, tiger_id

        return None

    def _check_range(
        self,
        segment: gpd.GeoSeries,
        house_number: int,
    ) -> Optional[Tuple[str, int, int]]:
        """
        Check if house number falls in segment's address range.

        Considers parity (odd/even) to determine correct side.

        Args:
            segment: Row from address features GeoDataFrame
            house_number: House number to check

        Returns:
            Tuple of (side, from_addr, to_addr) or None if not in range
        """
        # Check left side
        try:
            lfrom = int(segment["LFROMHN"]) if segment["LFROMHN"] else None
            lto = int(segment["LTOHN"]) if segment["LTOHN"] else None
        except (ValueError, TypeError):
            lfrom, lto = None, None

        if lfrom is not None and lto is not None:
            range_min, range_max = min(lfrom, lto), max(lfrom, lto)
            if range_min <= house_number <= range_max:
                # Check parity
                parity = segment.get("PARITYL", "B")
                if self._parity_matches(house_number, parity, lfrom):
                    return "L", lfrom, lto

        # Check right side
        try:
            rfrom = int(segment["RFROMHN"]) if segment["RFROMHN"] else None
            rto = int(segment["RTOHN"]) if segment["RTOHN"] else None
        except (ValueError, TypeError):
            rfrom, rto = None, None

        if rfrom is not None and rto is not None:
            range_min, range_max = min(rfrom, rto), max(rfrom, rto)
            if range_min <= house_number <= range_max:
                # Check parity
                parity = segment.get("PARITYR", "B")
                if self._parity_matches(house_number, parity, rfrom):
                    return "R", rfrom, rto

        return None

    def _parity_matches(
        self,
        house_number: int,
        parity: Optional[str],
        range_start: int,
    ) -> bool:
        """Check if house number matches the parity requirement."""
        if parity is None or parity == "B":
            return True

        house_is_odd = house_number % 2 == 1
        range_is_odd = range_start % 2 == 1

        if parity == "O":
            return house_is_odd
        elif parity == "E":
            return not house_is_odd
        else:
            # Fall back to matching range start parity
            return house_is_odd == range_is_odd

    def _interpolate_position(
        self,
        segment_geom: LineString,
        house_number: int,
        from_addr: int,
        to_addr: int,
        side: str,
        offset_meters: float = 10.0,
    ) -> Point:
        """
        Interpolate address position along segment.

        Args:
            segment_geom: Street segment LineString
            house_number: Target house number
            from_addr: Start of address range
            to_addr: End of address range
            side: "L" or "R" for left/right
            offset_meters: Distance to offset from centerline (not implemented)

        Returns:
            Point at interpolated position
        """
        # Calculate position along segment (0.0 to 1.0)
        if to_addr == from_addr:
            position = 0.5
        else:
            position = (house_number - from_addr) / (to_addr - from_addr)

        # Clamp to valid range
        position = max(0.0, min(1.0, position))

        # Interpolate point along line
        point = segment_geom.interpolate(position, normalized=True)

        # Note: Offsetting to the correct side of the street would require
        # projecting to a meter-based CRS, which adds complexity.
        # For most geocoding purposes, the centerline position is sufficient.

        return point



"""Data manager for orchestrating downloads, caching, and loading."""

import asyncio
import shutil
import uuid
from pathlib import Path
from typing import Dict, List, Optional, Set

import geopandas as gpd
import pandas as pd

from census_lookup.core.geoid import GeoLevel
from census_lookup.data.catalog import DataCatalog, DatasetInfo
from census_lookup.data.constants import FIPS_STATES, TIGER_URLS, normalize_state
from census_lookup.data.converter import GeoParquetConverter
from census_lookup.data.downloader import (
    ACSDataDownloader,
    CensusDataDownloader,
    TIGERDownloader,
    _coordinator,
)
from census_lookup.data.duckdb_engine import DuckDBEngine


class DataNotAvailableError(Exception):
    """Required data not downloaded."""

    def __init__(self, state: str, data_type: str):
        self.state = state
        self.data_type = data_type
        super().__init__(
            f"Data not available for {state} ({data_type}). "
            f"Run `census-lookup download {state}` to download."
        )


class DataManager:
    """
    Orchestrates data downloads, conversion, and caching.

    Directory structure:
    ~/.census-lookup/
    ├── catalog.json           # Tracks downloaded data
    ├── tiger/
    │   ├── addrfeat/          # Address range features
    │   │   ├── 06/            # State FIPS
    │   │   │   ├── 06001.parquet  # County-level files
    │   │   │   └── ...
    │   │   └── ...
    │   └── blocks/
    │       ├── 06.parquet     # State-level block files
    │       └── ...
    ├── census/
    │   ├── pl94171/
    │   │   ├── 06.parquet     # State-level census data
    │   │   └── ...
    │   └── variables.json     # Variable definitions
    └── temp/                   # Temporary download directory

    All download methods are async for concurrent operation.
    """

    def __init__(
        self,
        data_dir: Optional[Path] = None,
        auto_download: bool = True,
    ):
        """
        Initialize DataManager.

        Args:
            data_dir: Base directory for data. Defaults to ~/.census-lookup/
            auto_download: Whether to automatically download missing data
        """
        self.data_dir = data_dir or Path.home() / ".census-lookup"
        self.auto_download = auto_download

        # Create subdirectories
        self.tiger_dir = self.data_dir / "tiger"
        self.census_dir = self.data_dir / "census"
        self.temp_dir = self.data_dir / "temp"

        for d in [self.tiger_dir, self.census_dir, self.temp_dir]:
            d.mkdir(parents=True, exist_ok=True)

        # Initialize components
        self.catalog = DataCatalog(self.data_dir / "catalog.json")
        self.downloader = TIGERDownloader()
        self.census_downloader = CensusDataDownloader()
        self.acs_downloader = ACSDataDownloader()
        self.converter = GeoParquetConverter()

        # DuckDB engine (lazy initialization)
        self._duckdb: Optional[DuckDBEngine] = None

    @property
    def duckdb(self) -> DuckDBEngine:
        """Get or create DuckDB engine."""
        if self._duckdb is None:
            self._duckdb = DuckDBEngine(self.data_dir)
        return self._duckdb

    async def close(self):
        """Close all async sessions."""
        await self.downloader.close()
        await self.census_downloader.close()
        await self.acs_downloader.close()

    async def ensure_state_data(
        self,
        state: str,
        data_types: Optional[Set[str]] = None,
        show_progress: bool = True,
    ) -> None:
        """
        Ensure all required data for a state is downloaded and converted.

        Args:
            state: State name, abbreviation, or FIPS code
            data_types: Types of data to ensure. Defaults to {"blocks", "addrfeat", "pl94171"}
            show_progress: Show progress indicators
        """
        state_fips = normalize_state(state)

        if data_types is None:
            data_types = {"blocks", "addrfeat", "pl94171"}

        # Download all data types concurrently
        tasks = []
        if "blocks" in data_types:
            tasks.append(self._ensure_blocks(state_fips, show_progress))
        if "addrfeat" in data_types:
            tasks.append(self._ensure_address_features(state_fips, show_progress))
        if "pl94171" in data_types:
            tasks.append(self._ensure_census_data(state_fips, show_progress))

        await asyncio.gather(*tasks)

    async def _ensure_blocks(self, state_fips: str, show_progress: bool = True) -> None:
        """Ensure block data is available for a state."""
        if self.catalog.is_available("blocks", state_fips):
            return

        # Use coordinator to prevent concurrent downloads and conversions
        resource_key = f"ensure_blocks/{state_fips}"

        async def do_ensure():
            if self.catalog.is_available("blocks", state_fips):
                return self.catalog.get_path("blocks", state_fips)

            if show_progress:
                print(f"Downloading block data for {FIPS_STATES.get(state_fips, state_fips)}...")

            # Download shapefile
            temp_extract = await self.downloader.download_blocks(state_fips, self.temp_dir)

            # Convert to parquet
            output_path = self.tiger_dir / "blocks" / f"{state_fips}.parquet"
            self.converter.convert_blocks(temp_extract, output_path)

            # Register in catalog
            info = DatasetInfo.create(
                dataset_type="blocks",
                state_fips=state_fips,
                file_path=output_path,
                source_url=TIGER_URLS["blocks"].format(state_fips=state_fips),
            )
            self.catalog.register(info)

            # Clean up temp
            shutil.rmtree(temp_extract, ignore_errors=True)
            return output_path

        await _coordinator.download_once(resource_key, do_ensure)

    async def _ensure_address_features(self, state_fips: str, show_progress: bool = True) -> None:
        """Ensure address feature data is available for a state."""
        if self.catalog.is_available("addrfeat", state_fips):
            return

        # Use coordinator to prevent concurrent downloads and conversions
        resource_key = f"ensure_addrfeat/{state_fips}"

        async def do_ensure():
            if self.catalog.is_available("addrfeat", state_fips):
                return self.catalog.get_path("addrfeat", state_fips)

            if show_progress:
                state_name = FIPS_STATES.get(state_fips, state_fips)
                print(f"Downloading address features for {state_name}...")

            # Get list of counties for this state
            county_fips_list = self._get_county_fips_list(state_fips)

            # Download each county concurrently
            county_files = await self.downloader.download_address_features_for_state(
                state_fips,
                county_fips_list,
                self.temp_dir,
            )

            # Convert each to parquet
            parquet_files = []
            for shp_dir in county_files:
                county_fips = shp_dir.name.split("_")[2]  # Extract from tl_2020_XXXXX_addrfeat
                output_path = self.tiger_dir / "addrfeat" / state_fips / f"{county_fips}.parquet"
                self.converter.convert_address_features(shp_dir, output_path)
                parquet_files.append(output_path)

            # Merge into single state file
            state_output = self.tiger_dir / "addrfeat" / f"{state_fips}.parquet"
            if parquet_files:
                self.converter.merge_county_files(parquet_files, state_output)

                # Register in catalog
                info = DatasetInfo.create(
                    dataset_type="addrfeat",
                    state_fips=state_fips,
                    file_path=state_output,
                    source_url=TIGER_URLS["addrfeat"].format(county_fips="*"),
                )
                self.catalog.register(info)

            # Clean up temp
            for shp_dir in county_files:
                shutil.rmtree(shp_dir, ignore_errors=True)

            return state_output

        await _coordinator.download_once(resource_key, do_ensure)

    async def _ensure_census_data(self, state_fips: str, show_progress: bool = True) -> None:
        """Ensure PL 94-171 census data is available for a state."""
        if self.catalog.is_available("pl94171", state_fips):
            return

        # Use coordinator to prevent concurrent downloads and conversions
        resource_key = f"ensure_pl94171/{state_fips}"

        async def do_ensure():
            if self.catalog.is_available("pl94171", state_fips):
                return self.catalog.get_path("pl94171", state_fips)

            if show_progress:
                print(f"Downloading census data for {FIPS_STATES.get(state_fips, state_fips)}...")

            # Download via Census API
            from census_lookup.census.variables import DEFAULT_VARIABLES

            # Use unique temp file to avoid race conditions with concurrent downloads
            csv_path = self.temp_dir / f"pl94171_{state_fips}_{uuid.uuid4().hex[:8]}.csv"
            await self.census_downloader.download_pl94171_for_state(
                state_fips,
                variables=DEFAULT_VARIABLES,
                geo_level="block",
                dest_path=csv_path,
            )

            # Convert to parquet
            output_path = self.census_dir / "pl94171" / f"{state_fips}.parquet"
            self.converter.convert_census_csv(csv_path, output_path)

            # Register in catalog
            info = DatasetInfo.create(
                dataset_type="pl94171",
                state_fips=state_fips,
                file_path=output_path,
                source_url="https://api.census.gov/data/2020/dec/pl",
            )
            self.catalog.register(info)

            # Clean up
            csv_path.unlink(missing_ok=True)
            return output_path

        await _coordinator.download_once(resource_key, do_ensure)

    def _get_county_fips_list(self, state_fips: str) -> List[str]:
        """
        Get list of county FIPS codes for a state.

        Uses static county data from constants module.
        """
        from census_lookup.data.constants import get_counties_for_state

        counties = get_counties_for_state(state_fips)
        # Return full 5-digit county FIPS (state + county)
        return [f"{state_fips}{county}" for county in counties]

    async def get_blocks(self, state_fips: str) -> gpd.GeoDataFrame:
        """
        Load block polygons for a state.

        Args:
            state_fips: 2-digit state FIPS code

        Returns:
            GeoDataFrame with block polygons
        """
        state_fips = normalize_state(state_fips)

        if self.auto_download and not self.catalog.is_available("blocks", state_fips):
            await self._ensure_blocks(state_fips)

        path = self.catalog.get_path("blocks", state_fips)
        if not path:
            raise DataNotAvailableError(state_fips, "blocks")

        return gpd.read_parquet(path)

    async def get_address_features(self, state_fips: str) -> gpd.GeoDataFrame:
        """
        Load address features for a state.

        Args:
            state_fips: 2-digit state FIPS code

        Returns:
            GeoDataFrame with address range features
        """
        state_fips = normalize_state(state_fips)

        if self.auto_download and not self.catalog.is_available("addrfeat", state_fips):
            await self._ensure_address_features(state_fips)

        path = self.catalog.get_path("addrfeat", state_fips)
        if not path:
            raise DataNotAvailableError(state_fips, "addrfeat")

        return gpd.read_parquet(path)

    def clear_cache(self, state: Optional[str] = None) -> None:
        """
        Clear cached data.

        Args:
            state: State to clear (None for all states)
        """
        if state:
            state_fips = normalize_state(state)

            # Remove files
            for dataset_type in ["blocks", "addrfeat", "pl94171"]:
                path = self.catalog.get_path(dataset_type, state_fips)
                if path and path.exists():
                    path.unlink()
                self.catalog.unregister(dataset_type, state_fips)
        else:
            # Clear all
            shutil.rmtree(self.tiger_dir, ignore_errors=True)
            shutil.rmtree(self.census_dir, ignore_errors=True)
            self.catalog.clear()

            # Recreate directories
            self.tiger_dir.mkdir(parents=True, exist_ok=True)
            self.census_dir.mkdir(parents=True, exist_ok=True)

    def disk_usage(self) -> Dict[str, int]:
        """
        Get disk usage by category.

        Returns:
            Dictionary with category names and sizes in bytes
        """
        usage = {"total": 0}

        for category in ["tiger/blocks", "tiger/addrfeat", "census/pl94171"]:
            path = self.data_dir / category
            if path.exists():
                size = sum(f.stat().st_size for f in path.rglob("*") if f.is_file())
                usage[category] = size
                usage["total"] += size

        return usage

    def list_available_states(self, dataset_type: str = "blocks") -> List[str]:
        """List states with available data."""
        return self.catalog.list_states(dataset_type)

    # ==========================================================================
    # ACS Data Support
    # ==========================================================================

    async def ensure_acs_data(
        self,
        state: str,
        variables: Optional[List[str]] = None,
        geo_level: str = "tract",
        show_progress: bool = True,
    ) -> None:
        """
        Ensure ACS data is downloaded for a state.

        Args:
            state: State name, abbreviation, or FIPS code
            variables: ACS variables to download (defaults to DEFAULT_ACS_VARIABLES)
            geo_level: Geographic level (tract, block group, county)
            show_progress: Show progress indicators
        """
        state_fips = normalize_state(state)
        await self._ensure_acs_data(state_fips, variables, geo_level, show_progress)

    async def _ensure_acs_data(
        self,
        state_fips: str,
        variables: Optional[List[str]] = None,
        geo_level: str = "tract",
        show_progress: bool = True,
    ) -> None:
        """Ensure ACS data is available for a state."""
        # Check if already downloaded
        dataset_key = f"acs5_{geo_level}"
        if self.catalog.is_available(dataset_key, state_fips):
            return

        # Use coordinator to prevent concurrent downloads and conversions
        resource_key = f"ensure_acs5/{state_fips}/{geo_level}"

        async def do_ensure():
            if self.catalog.is_available(dataset_key, state_fips):
                return self.catalog.get_path(dataset_key, state_fips)

            if show_progress:
                print(
                    f"Downloading ACS data for {FIPS_STATES.get(state_fips, state_fips)} "
                    f"at {geo_level} level..."
                )

            # Get default variables if not specified
            vars_to_use = variables
            if vars_to_use is None:
                from census_lookup.census.acs import DEFAULT_ACS_VARIABLES

                vars_to_use = DEFAULT_ACS_VARIABLES

            # Download via Census API
            # Use unique temp file to avoid race conditions with concurrent downloads
            csv_path = self.temp_dir / f"acs5_{state_fips}_{geo_level}_{uuid.uuid4().hex[:8]}.csv"
            await self.acs_downloader.download_acs_for_state(
                state_fips,
                variables=vars_to_use,
                geo_level=geo_level,
                dest_path=csv_path,
            )

            # Convert to parquet
            output_path = self.census_dir / "acs5" / geo_level / f"{state_fips}.parquet"
            self.converter.convert_census_csv(csv_path, output_path)

            # Register in catalog
            info = DatasetInfo.create(
                dataset_type=dataset_key,
                state_fips=state_fips,
                file_path=output_path,
                source_url=self.acs_downloader.api_base,
            )
            self.catalog.register(info)

            # Clean up
            csv_path.unlink(missing_ok=True)
            return output_path

        await _coordinator.download_once(resource_key, do_ensure)

    async def get_acs_data(
        self,
        state_fips: str,
        variables: List[str],
        geo_level: GeoLevel = GeoLevel.TRACT,
    ) -> pd.DataFrame:
        """
        Load ACS data for specified variables.

        Args:
            state_fips: 2-digit state FIPS code
            variables: ACS variable codes
            geo_level: Geographic level (TRACT, BLOCK_GROUP, or COUNTY)

        Returns:
            DataFrame with GEOID and requested variables
        """
        state_fips = normalize_state(state_fips)

        # Map GeoLevel to string for catalog
        geo_level_str = {
            GeoLevel.TRACT: "tract",
            GeoLevel.BLOCK_GROUP: "block group",
            GeoLevel.COUNTY: "county",
        }.get(geo_level)

        if geo_level_str is None:
            raise ValueError(
                f"ACS data is not available at {geo_level.name} level. "
                "Use TRACT, BLOCK_GROUP, or COUNTY."
            )

        dataset_key = f"acs5_{geo_level_str}"

        if self.auto_download and not self.catalog.is_available(dataset_key, state_fips):
            await self._ensure_acs_data(state_fips, variables, geo_level_str)

        path = self.catalog.get_path(dataset_key, state_fips)
        if not path:
            raise DataNotAvailableError(state_fips, dataset_key)

        # Use DuckDB for efficient querying
        var_list = ", ".join(variables)

        # Read available columns and select those that exist
        check_sql = f"SELECT * FROM read_parquet('{path}') LIMIT 0"
        available_cols = set(self.duckdb.query(check_sql).columns)

        # Filter to only variables that exist in the data
        existing_vars = [v for v in variables if v in available_cols]
        if not existing_vars:
            raise ValueError(
                f"None of the requested variables found in ACS data. "
                f"Available columns: {available_cols}"
            )

        var_list = ", ".join(existing_vars)
        sql = f"SELECT GEOID, {var_list} FROM read_parquet('{path}')"

        return self.duckdb.query(sql)



"""American Community Survey (ACS) 5-Year Estimates variable definitions.

ACS provides richer demographic and socioeconomic data beyond PL 94-171,
including income, education, employment, housing characteristics, and more.

Reference: https://www.census.gov/data/developers/data-sets/acs-5year.html
"""

from typing import Dict, List

# ACS 5-Year Variable Definitions (2020)
# Format: Table_ColumnE for estimates, Table_ColumnM for margins of error
# Reference: https://api.census.gov/data/2020/acs/acs5/variables.html

ACS_VARIABLES: Dict[str, str] = {
    # Demographics (B01)
    "B01001_001E": "Total Population",
    "B01001_002E": "Male Population",
    "B01001_026E": "Female Population",
    "B01002_001E": "Median Age",
    "B01003_001E": "Total Population (alternate)",

    # Age Distribution (B01001)
    "B01001_003E": "Male Under 5 years",
    "B01001_004E": "Male 5 to 9 years",
    "B01001_005E": "Male 10 to 14 years",
    "B01001_006E": "Male 15 to 17 years",
    "B01001_007E": "Male 18 and 19 years",
    "B01001_008E": "Male 20 years",
    "B01001_009E": "Male 21 years",
    "B01001_010E": "Male 22 to 24 years",
    "B01001_011E": "Male 25 to 29 years",
    "B01001_012E": "Male 30 to 34 years",
    "B01001_013E": "Male 35 to 39 years",
    "B01001_014E": "Male 40 to 44 years",
    "B01001_015E": "Male 45 to 49 years",
    "B01001_016E": "Male 50 to 54 years",
    "B01001_017E": "Male 55 to 59 years",
    "B01001_018E": "Male 60 and 61 years",
    "B01001_019E": "Male 62 to 64 years",
    "B01001_020E": "Male 65 and 66 years",
    "B01001_021E": "Male 67 to 69 years",
    "B01001_022E": "Male 70 to 74 years",
    "B01001_023E": "Male 75 to 79 years",
    "B01001_024E": "Male 80 to 84 years",
    "B01001_025E": "Male 85 years and over",

    # Income (B19)
    "B19013_001E": "Median Household Income",
    "B19001_001E": "Households by Income - Total",
    "B19001_002E": "Households Income Less than $10,000",
    "B19001_003E": "Households Income $10,000 to $14,999",
    "B19001_004E": "Households Income $15,000 to $19,999",
    "B19001_005E": "Households Income $20,000 to $24,999",
    "B19001_006E": "Households Income $25,000 to $29,999",
    "B19001_007E": "Households Income $30,000 to $34,999",
    "B19001_008E": "Households Income $35,000 to $39,999",
    "B19001_009E": "Households Income $40,000 to $44,999",
    "B19001_010E": "Households Income $45,000 to $49,999",
    "B19001_011E": "Households Income $50,000 to $59,999",
    "B19001_012E": "Households Income $60,000 to $74,999",
    "B19001_013E": "Households Income $75,000 to $99,999",
    "B19001_014E": "Households Income $100,000 to $124,999",
    "B19001_015E": "Households Income $125,000 to $149,999",
    "B19001_016E": "Households Income $150,000 to $199,999",
    "B19001_017E": "Households Income $200,000 or more",
    "B19301_001E": "Per Capita Income",
    "B19083_001E": "Gini Index of Income Inequality",

    # Poverty (B17)
    "B17001_001E": "Poverty Status - Total Population",
    "B17001_002E": "Below Poverty Level",
    "B17001_031E": "At or Above Poverty Level",
    "B17020_001E": "Poverty Status by Age - Total",
    "B17020_002E": "Poverty Status Under 6 years",

    # Education (B15)
    "B15003_001E": "Educational Attainment - Population 25+",
    "B15003_002E": "No schooling completed",
    "B15003_003E": "Nursery school",
    "B15003_004E": "Kindergarten",
    "B15003_005E": "1st grade",
    "B15003_006E": "2nd grade",
    "B15003_007E": "3rd grade",
    "B15003_008E": "4th grade",
    "B15003_009E": "5th grade",
    "B15003_010E": "6th grade",
    "B15003_011E": "7th grade",
    "B15003_012E": "8th grade",
    "B15003_013E": "9th grade",
    "B15003_014E": "10th grade",
    "B15003_015E": "11th grade",
    "B15003_016E": "12th grade, no diploma",
    "B15003_017E": "High school diploma",
    "B15003_018E": "GED or alternative credential",
    "B15003_019E": "Some college, less than 1 year",
    "B15003_020E": "Some college, 1 or more years, no degree",
    "B15003_021E": "Associate's degree",
    "B15003_022E": "Bachelor's degree",
    "B15003_023E": "Master's degree",
    "B15003_024E": "Professional school degree",
    "B15003_025E": "Doctorate degree",

    # Employment (B23)
    "B23025_001E": "Employment Status - Population 16+",
    "B23025_002E": "In Labor Force",
    "B23025_003E": "Civilian Labor Force",
    "B23025_004E": "Employed",
    "B23025_005E": "Unemployed",
    "B23025_006E": "Armed Forces",
    "B23025_007E": "Not in Labor Force",

    # Occupation (C24)
    "C24010_001E": "Occupation - Civilian employed 16+",
    "C24010_002E": "Male Civilian employed 16+",
    "C24010_003E": "Male Management, business, science, arts",
    "C24010_038E": "Female Civilian employed 16+",
    "C24010_039E": "Female Management, business, science, arts",

    # Industry (C24)
    "C24030_001E": "Industry - Civilian employed 16+",
    "C24030_002E": "Agriculture, forestry, fishing, hunting, mining",
    "C24030_003E": "Construction",
    "C24030_004E": "Manufacturing",
    "C24030_005E": "Wholesale trade",
    "C24030_006E": "Retail trade",
    "C24030_007E": "Transportation, warehousing, utilities",
    "C24030_008E": "Information",
    "C24030_009E": "Finance, insurance, real estate",
    "C24030_010E": "Professional, scientific, management, admin, waste",
    "C24030_011E": "Educational services, health care, social assistance",
    "C24030_012E": "Arts, entertainment, recreation, food services",
    "C24030_013E": "Other services",
    "C24030_014E": "Public administration",

    # Commute (B08)
    "B08301_001E": "Means of Transportation to Work - Total",
    "B08301_002E": "Car, truck, or van",
    "B08301_003E": "Car, truck, van - drove alone",
    "B08301_004E": "Car, truck, van - carpooled",
    "B08301_010E": "Public transportation",
    "B08301_016E": "Taxicab",
    "B08301_017E": "Motorcycle",
    "B08301_018E": "Bicycle",
    "B08301_019E": "Walked",
    "B08301_020E": "Other means",
    "B08301_021E": "Worked from home",
    "B08303_001E": "Travel Time to Work - Total",
    "B08013_001E": "Aggregate Travel Time to Work (minutes)",

    # Housing (B25)
    "B25001_001E": "Total Housing Units",
    "B25002_001E": "Occupancy Status - Total",
    "B25002_002E": "Occupied Housing Units",
    "B25002_003E": "Vacant Housing Units",
    "B25003_001E": "Tenure - Total Occupied Units",
    "B25003_002E": "Owner-Occupied",
    "B25003_003E": "Renter-Occupied",
    "B25024_001E": "Units in Structure - Total",
    "B25024_002E": "1 unit, detached",
    "B25024_003E": "1 unit, attached",
    "B25024_004E": "2 units",
    "B25024_005E": "3 or 4 units",
    "B25024_006E": "5 to 9 units",
    "B25024_007E": "10 to 19 units",
    "B25024_008E": "20 to 49 units",
    "B25024_009E": "50 or more units",
    "B25024_010E": "Mobile home",
    "B25024_011E": "Boat, RV, van, etc.",
    "B25035_001E": "Median Year Structure Built",
    "B25064_001E": "Median Gross Rent",
    "B25077_001E": "Median Home Value",
    "B25071_001E": "Median Gross Rent as % of Household Income",
    "B25070_001E": "Gross Rent as % of Income - Total",
    "B25070_010E": "Gross Rent 50% or more of income",

    # Health Insurance (B27)
    "B27001_001E": "Health Insurance Coverage - Total",
    "B27001_004E": "Under 6 with health insurance",
    "B27001_005E": "Under 6 without health insurance",
    "B27010_001E": "Health Insurance by Type - Total",
    "B27010_017E": "Employer-based insurance",
    "B27010_033E": "Direct-purchase insurance",
    "B27010_050E": "Medicare",
    "B27010_066E": "Medicaid/means-tested public coverage",

    # Household Composition (B11)
    "B11001_001E": "Households - Total",
    "B11001_002E": "Family households",
    "B11001_003E": "Married-couple family",
    "B11001_004E": "Male householder, no spouse",
    "B11001_005E": "Female householder, no spouse",
    "B11001_007E": "Nonfamily households",
    "B11001_008E": "Householder living alone",
    "B11016_001E": "Household Type by Size - Total",
    "B25010_001E": "Average Household Size",

    # Language (B16)
    "B16001_001E": "Language Spoken at Home - Population 5+",
    "B16001_002E": "Speak only English",
    "B16001_003E": "Speak Spanish",
    "B16001_006E": "Speak French, Haitian, or Cajun",
    "B16001_009E": "Speak German or West Germanic",
    "B16001_012E": "Speak Russian, Polish, or other Slavic",
    "B16001_015E": "Speak other Indo-European",
    "B16001_018E": "Speak Korean",
    "B16001_021E": "Speak Chinese (incl. Mandarin, Cantonese)",
    "B16001_024E": "Speak Vietnamese",
    "B16001_027E": "Speak Tagalog",
    "B16001_030E": "Speak other Asian/Pacific Islander",
    "B16001_033E": "Speak Arabic",
    "B16001_036E": "Speak other/unspecified language",

    # Nativity and Citizenship (B05)
    "B05001_001E": "Nativity and Citizenship - Total",
    "B05001_002E": "U.S. citizen, born in US",
    "B05001_003E": "U.S. citizen, born in PR or Island Areas",
    "B05001_004E": "U.S. citizen, born abroad of American parents",
    "B05001_005E": "U.S. citizen by naturalization",
    "B05001_006E": "Not a U.S. citizen",

    # Internet Access (B28)
    "B28002_001E": "Internet Access - Total Households",
    "B28002_002E": "With an Internet subscription",
    "B28002_003E": "Dial-up with no other type",
    "B28002_004E": "Broadband of any type",
    "B28002_007E": "Cellular data plan",
    "B28002_012E": "Without Internet subscription",
    "B28002_013E": "No Internet access",
    "B28003_001E": "Computer in Household - Total",
    "B28003_002E": "Has a computer",
    "B28003_004E": "No computer",

    # Vehicles Available (B25044)
    "B25044_001E": "Vehicles Available - Occupied Housing Units",
    "B25044_003E": "Owner-occupied - No vehicle",
    "B25044_004E": "Owner-occupied - 1 vehicle",
    "B25044_005E": "Owner-occupied - 2 vehicles",
    "B25044_006E": "Owner-occupied - 3 vehicles",
    "B25044_007E": "Owner-occupied - 4 vehicles",
    "B25044_008E": "Owner-occupied - 5+ vehicles",
    "B25044_010E": "Renter-occupied - No vehicle",
    "B25044_011E": "Renter-occupied - 1 vehicle",
}

# Common variable groups for convenience
ACS_VARIABLE_GROUPS: Dict[str, List[str]] = {
    "demographics": [
        "B01001_001E",  # Total population
        "B01001_002E",  # Male
        "B01001_026E",  # Female
        "B01002_001E",  # Median age
    ],
    "income": [
        "B19013_001E",  # Median household income
        "B19301_001E",  # Per capita income
        "B19083_001E",  # Gini index
    ],
    "income_distribution": [f"B19001_{i:03d}E" for i in range(1, 18)],
    "poverty": [
        "B17001_001E",  # Total
        "B17001_002E",  # Below poverty
        "B17001_031E",  # At or above poverty
    ],
    "education": [
        "B15003_001E",  # Total 25+
        "B15003_017E",  # HS diploma
        "B15003_021E",  # Associate's
        "B15003_022E",  # Bachelor's
        "B15003_023E",  # Master's
        "B15003_024E",  # Professional
        "B15003_025E",  # Doctorate
    ],
    "education_detailed": [f"B15003_{i:03d}E" for i in range(1, 26)],
    "employment": [
        "B23025_001E",  # Total 16+
        "B23025_002E",  # In labor force
        "B23025_004E",  # Employed
        "B23025_005E",  # Unemployed
        "B23025_007E",  # Not in labor force
    ],
    "commute": [
        "B08301_001E",  # Total workers
        "B08301_003E",  # Drove alone
        "B08301_004E",  # Carpooled
        "B08301_010E",  # Public transit
        "B08301_019E",  # Walked
        "B08301_021E",  # Worked from home
    ],
    "housing": [
        "B25001_001E",  # Total units
        "B25002_002E",  # Occupied
        "B25002_003E",  # Vacant
        "B25003_002E",  # Owner-occupied
        "B25003_003E",  # Renter-occupied
        "B25077_001E",  # Median home value
        "B25064_001E",  # Median gross rent
    ],
    "housing_detailed": [
        "B25001_001E",
        "B25002_002E",
        "B25002_003E",
        "B25003_002E",
        "B25003_003E",
        "B25077_001E",
        "B25064_001E",
        "B25035_001E",  # Median year built
        "B25071_001E",  # Rent burden
    ] + [f"B25024_{i:03d}E" for i in range(2, 12)],  # Units in structure
    "health_insurance": [
        "B27001_001E",  # Total
        "B27010_017E",  # Employer-based
        "B27010_033E",  # Direct-purchase
        "B27010_050E",  # Medicare
        "B27010_066E",  # Medicaid
    ],
    "household": [
        "B11001_001E",  # Total households
        "B11001_002E",  # Family
        "B11001_003E",  # Married-couple
        "B11001_007E",  # Nonfamily
        "B25010_001E",  # Avg household size
    ],
    "language": [
        "B16001_001E",  # Total 5+
        "B16001_002E",  # English only
        "B16001_003E",  # Spanish
        "B16001_021E",  # Chinese
    ],
    "citizenship": [
        "B05001_001E",  # Total
        "B05001_002E",  # Born in US
        "B05001_005E",  # Naturalized
        "B05001_006E",  # Not citizen
    ],
    "internet": [
        "B28002_001E",  # Total households
        "B28002_004E",  # Broadband
        "B28002_007E",  # Cellular
        "B28002_013E",  # No internet
    ],
    "vehicles": [
        "B25044_001E",  # Total occupied
        "B25044_003E",  # Owner - no vehicle
        "B25044_010E",  # Renter - no vehicle
    ],
}

# Default ACS variables (commonly used subset)
DEFAULT_ACS_VARIABLES: List[str] = [
    "B01001_001E",  # Total population
    "B01002_001E",  # Median age
    "B19013_001E",  # Median household income
    "B19301_001E",  # Per capita income
    "B17001_002E",  # Below poverty level
    "B15003_022E",  # Bachelor's degree
    "B23025_004E",  # Employed
    "B23025_005E",  # Unemployed
    "B25001_001E",  # Total housing units
    "B25077_001E",  # Median home value
    "B25064_001E",  # Median gross rent
    "B25003_002E",  # Owner-occupied
    "B25003_003E",  # Renter-occupied
    "B28002_004E",  # Broadband internet
]


def get_acs_variables_for_group(group: str) -> List[str]:
    """
    Get list of ACS variables for a named group.

    Args:
        group: Group name (e.g., "income", "education", "housing")

    Returns:
        List of variable codes

    Raises:
        ValueError: If group name is not recognized
    """
    if group not in ACS_VARIABLE_GROUPS:
        valid = ", ".join(ACS_VARIABLE_GROUPS.keys())
        raise ValueError(f"Unknown ACS variable group: {group}. Valid groups: {valid}")

    return ACS_VARIABLE_GROUPS[group]


def list_acs_tables() -> Dict[str, str]:
    """List available ACS table groups."""
    return {
        "B01": "Sex and Age",
        "B05": "Nativity and Citizenship",
        "B08": "Commuting/Transportation",
        "B11": "Household Type and Relationships",
        "B15": "Educational Attainment",
        "B16": "Language Spoken at Home",
        "B17": "Poverty Status",
        "B19": "Income",
        "B23": "Employment Status",
        "B25": "Housing Characteristics",
        "B27": "Health Insurance",
        "B28": "Internet Access and Computers",
        "C24": "Industry and Occupation",
    }


def list_acs_variable_groups() -> Dict[str, str]:
    """List available ACS variable groups with descriptions."""
    return {
        "demographics": "Age, sex, median age",
        "income": "Household income, per capita income, Gini index",
        "income_distribution": "Household income brackets",
        "poverty": "Poverty status",
        "education": "Educational attainment (common levels)",
        "education_detailed": "Educational attainment (all levels)",
        "employment": "Employment status and labor force",
        "commute": "Means of transportation to work",
        "housing": "Housing occupancy, tenure, values",
        "housing_detailed": "Housing characteristics (extended)",
        "health_insurance": "Health insurance coverage by type",
        "household": "Household composition and size",
        "language": "Language spoken at home",
        "citizenship": "Nativity and citizenship status",
        "internet": "Internet access and computer ownership",
        "vehicles": "Vehicles available per household",
    }



"""GEOID parsing and manipulation utilities."""

from dataclasses import dataclass
from enum import Enum


class GeoLevel(Enum):
    """Geographic hierarchy levels."""

    STATE = "state"
    COUNTY = "county"
    TRACT = "tract"
    BLOCK_GROUP = "block_group"
    BLOCK = "block"

    @property
    def geoid_length(self) -> int:
        """Return the GEOID length for this level."""
        lengths = {
            GeoLevel.STATE: 2,
            GeoLevel.COUNTY: 5,
            GeoLevel.TRACT: 11,
            GeoLevel.BLOCK_GROUP: 12,
            GeoLevel.BLOCK: 15,
        }
        return lengths[self]

    @classmethod
    def from_geoid_length(cls, length: int) -> "GeoLevel":
        """Determine GeoLevel from GEOID length."""
        if length >= 15:
            return cls.BLOCK
        elif length >= 12:
            return cls.BLOCK_GROUP
        elif length >= 11:
            return cls.TRACT
        elif length >= 5:
            return cls.COUNTY
        else:
            return cls.STATE


@dataclass
class GEOIDComponents:
    """Parsed GEOID components.

    Always created from a full block GEOID (15 digits), so all components
    are guaranteed to be set.
    """

    state: str  # 2 digits
    county: str  # 3 digits
    tract: str  # 6 digits
    block_group: str  # 1 digit
    block: str  # 4 digits (includes block_group)

    @property
    def county_fips(self) -> str:
        """Return full county FIPS code (state + county)."""
        return self.state + self.county

    @property
    def tract_geoid(self) -> str:
        """Return tract GEOID."""
        return self.state + self.county + self.tract

    @property
    def block_group_geoid(self) -> str:
        """Return block group GEOID."""
        return self.state + self.county + self.tract + self.block_group


class GEOIDParser:
    """
    Parse and manipulate GEOIDs.

    GEOID structure:
    - State: 2 digits (e.g., "06" for California)
    - County: 3 digits (e.g., "037" for Los Angeles)
    - Tract: 6 digits (e.g., "101100")
    - Block Group: 1 digit (e.g., "1")
    - Block: 4 digits (e.g., "1001")

    Full block GEOID: 15 digits (e.g., "060371011001001")
    """

    @staticmethod
    def parse(geoid: str) -> GEOIDComponents:
        """
        Parse a full block GEOID (15 digits) into components.

        The GEOID is validated at data load time (in converter.py), so this
        method assumes valid input.

        Args:
            geoid: A 15-digit block GEOID string

        Returns:
            GEOIDComponents with all values set
        """
        return GEOIDComponents(
            state=geoid[:2],
            county=geoid[2:5],
            tract=geoid[5:11],
            block_group=geoid[11:12],
            block=geoid[11:15],
        )



"""Census 2020 PL 94-171 variable definitions."""

from typing import Dict, List

# PL 94-171 Variable Definitions
# Reference: https://www.census.gov/programs-surveys/decennial-census/about/rdo/summary-files.html

VARIABLES: Dict[str, str] = {
    # Table P1: Race
    "P1_001N": "Total Population",
    "P1_002N": "Population of one race",
    "P1_003N": "White alone",
    "P1_004N": "Black or African American alone",
    "P1_005N": "American Indian and Alaska Native alone",
    "P1_006N": "Asian alone",
    "P1_007N": "Native Hawaiian and Other Pacific Islander alone",
    "P1_008N": "Some Other Race alone",
    "P1_009N": "Population of two or more races",
    "P1_010N": "Population of two races",
    "P1_011N": "White; Black or African American",
    "P1_012N": "White; American Indian and Alaska Native",
    "P1_013N": "White; Asian",
    "P1_014N": "White; Native Hawaiian and Other Pacific Islander",
    "P1_015N": "White; Some Other Race",
    "P1_016N": "Black or African American; American Indian and Alaska Native",
    "P1_017N": "Black or African American; Asian",
    "P1_018N": "Black or African American; Native Hawaiian and Other Pacific Islander",
    "P1_019N": "Black or African American; Some Other Race",
    "P1_020N": "American Indian and Alaska Native; Asian",
    "P1_021N": "American Indian and Alaska Native; Native Hawaiian and Other Pacific Islander",
    "P1_022N": "American Indian and Alaska Native; Some Other Race",
    "P1_023N": "Asian; Native Hawaiian and Other Pacific Islander",
    "P1_024N": "Asian; Some Other Race",
    "P1_025N": "Native Hawaiian and Other Pacific Islander; Some Other Race",
    # Table P2: Hispanic or Latino by Race
    "P2_001N": "Total Population",
    "P2_002N": "Hispanic or Latino",
    "P2_003N": "Not Hispanic or Latino",
    "P2_004N": "Not Hispanic or Latino: Population of one race",
    "P2_005N": "Not Hispanic or Latino: White alone",
    "P2_006N": "Not Hispanic or Latino: Black or African American alone",
    "P2_007N": "Not Hispanic or Latino: American Indian and Alaska Native alone",
    "P2_008N": "Not Hispanic or Latino: Asian alone",
    "P2_009N": "Not Hispanic or Latino: Native Hawaiian and Other Pacific Islander alone",
    "P2_010N": "Not Hispanic or Latino: Some Other Race alone",
    "P2_011N": "Not Hispanic or Latino: Population of two or more races",
    # Table P3: Race for Population 18 Years and Over
    "P3_001N": "Total Population 18 years and over",
    "P3_002N": "Population 18+ of one race",
    "P3_003N": "Population 18+ White alone",
    "P3_004N": "Population 18+ Black or African American alone",
    "P3_005N": "Population 18+ American Indian and Alaska Native alone",
    "P3_006N": "Population 18+ Asian alone",
    "P3_007N": "Population 18+ Native Hawaiian and Other Pacific Islander alone",
    "P3_008N": "Population 18+ Some Other Race alone",
    "P3_009N": "Population 18+ of two or more races",
    # Table P4: Hispanic or Latino by Race for Population 18+
    "P4_001N": "Total Population 18 years and over",
    "P4_002N": "Hispanic or Latino 18+",
    "P4_003N": "Not Hispanic or Latino 18+",
    # Table H1: Housing Units
    "H1_001N": "Total Housing Units",
    "H1_002N": "Occupied Housing Units",
    "H1_003N": "Vacant Housing Units",
}

# Common variable groups for convenience
VARIABLE_GROUPS: Dict[str, List[str]] = {
    "population": ["P1_001N"],
    "race_simple": [
        "P1_001N",
        "P1_003N",
        "P1_004N",
        "P1_005N",
        "P1_006N",
        "P1_007N",
        "P1_008N",
    ],
    "race_detailed": [f"P1_{i:03d}N" for i in range(1, 26)],
    "hispanic": ["P2_001N", "P2_002N", "P2_003N"],
    "hispanic_detailed": [f"P2_{i:03d}N" for i in range(1, 12)],
    "voting_age": ["P3_001N"],
    "voting_age_race": [f"P3_{i:03d}N" for i in range(1, 10)],
    "housing": ["H1_001N", "H1_002N", "H1_003N"],
    "all": list(VARIABLES.keys()),
}

# Default variables to download (balance between completeness and size)
DEFAULT_VARIABLES: List[str] = [
    "P1_001N",  # Total population
    "P1_003N",  # White alone
    "P1_004N",  # Black alone
    "P1_005N",  # American Indian alone
    "P1_006N",  # Asian alone
    "P1_007N",  # Pacific Islander alone
    "P1_008N",  # Other race alone
    "P2_002N",  # Hispanic or Latino
    "P2_005N",  # Non-Hispanic White alone
    "P3_001N",  # Voting age population
    "H1_001N",  # Total housing units
    "H1_002N",  # Occupied housing units
    "H1_003N",  # Vacant housing units
]


def get_variables_for_group(group: str) -> List[str]:
    """
    Get list of variables for a named group.

    Args:
        group: Group name (e.g., "population", "race_simple", "housing")

    Returns:
        List of variable codes

    Raises:
        ValueError: If group name is not recognized
    """
    if group not in VARIABLE_GROUPS:
        valid = ", ".join(VARIABLE_GROUPS.keys())
        raise ValueError(f"Unknown variable group: {group}. Valid groups: {valid}")

    return VARIABLE_GROUPS[group]


def list_tables() -> Dict[str, str]:
    """List available PL 94-171 tables."""
    return {
        "P1": "Race",
        "P2": "Hispanic or Latino by Race",
        "P3": "Race for Population 18 Years and Over",
        "P4": "Hispanic or Latino by Race for Population 18+",
        "H1": "Housing Units",
    }


def list_variable_groups() -> Dict[str, str]:
    """List available variable groups with descriptions."""
    return {
        "population": "Total population only",
        "race_simple": "Population by major race categories",
        "race_detailed": "Population by all race combinations",
        "hispanic": "Hispanic/Latino population",
        "hispanic_detailed": "Hispanic/Latino by race",
        "voting_age": "Population 18 years and over",
        "voting_age_race": "Voting age population by race",
        "housing": "Housing unit counts",
        "all": "All available variables",
    }



"""DuckDB-based data engine for census queries and joins."""

from pathlib import Path
from typing import Dict, List, Optional, Union

import duckdb
import pandas as pd

from census_lookup.core.geoid import GeoLevel


class DuckDBEngine:
    """
    DuckDB-based data engine for census queries.

    Handles:
    - Querying GeoParquet/Parquet files directly
    - Joining census data to GEOIDs
    - Aggregating data to higher geographic levels
    - Efficient batch operations
    """

    def __init__(self, data_dir: Path):
        """
        Initialize DuckDB engine.

        Args:
            data_dir: Base directory containing census data files
        """
        self.data_dir = data_dir
        self.conn = duckdb.connect()
        self._setup_extensions()

    def _setup_extensions(self) -> None:
        """Install and load required DuckDB extensions."""
        # Spatial extension for geometry support
        self.conn.execute("INSTALL spatial")
        self.conn.execute("LOAD spatial")

    def close(self) -> None:
        """Close the DuckDB connection."""
        self.conn.close()

    def __enter__(self) -> "DuckDBEngine":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def query(self, sql: str, params: Optional[dict] = None) -> pd.DataFrame:
        """
        Execute a SQL query and return results as DataFrame.

        Args:
            sql: SQL query string
            params: Optional parameters for parameterized queries

        Returns:
            DataFrame with query results
        """
        if params:
            result = self.conn.execute(sql, params)
        else:
            result = self.conn.execute(sql)
        return result.fetchdf()

    def get_census_parquet_path(self, state_fips: str) -> Path:
        """Get the path to census data parquet file for a state."""
        return self.data_dir / "census" / "pl94171" / f"{state_fips}.parquet"

    def join_census_data(
        self,
        geoids: Union[List[str], pd.Series],
        variables: List[str],
        geo_level: GeoLevel = GeoLevel.BLOCK,
    ) -> pd.DataFrame:
        """
        Join census variables to a list of GEOIDs.

        Args:
            geoids: List or Series of GEOIDs
            variables: Census variable codes to retrieve
            geo_level: Geographic level of the GEOIDs

        Returns:
            DataFrame with GEOID and requested variables
        """
        if isinstance(geoids, pd.Series):
            geoids = geoids.tolist()

        if not geoids:
            return pd.DataFrame(columns=["GEOID"] + variables)

        # Determine which states we need
        state_fips_set = set(g[:2] for g in geoids if g and len(g) >= 2)

        # Build query for each state and union
        parquet_paths = []
        for state_fips in state_fips_set:
            path = self.get_census_parquet_path(state_fips)
            if path.exists():
                parquet_paths.append(str(path))

        if not parquet_paths:
            # No data available, return empty with GEOIDs
            return pd.DataFrame({"GEOID": geoids, **{v: None for v in variables}})

        # Register GEOIDs as a temporary table
        geoid_df = pd.DataFrame({"GEOID": geoids})
        self.conn.register("input_geoids", geoid_df)

        # Handle GEOID truncation for higher levels
        geoid_length = geo_level.geoid_length

        # Build parquet glob pattern
        parquet_glob = f"read_parquet([{', '.join(repr(p) for p in parquet_paths)}])"

        # Build the variable list for SELECT (from the subquery result 'c')
        var_list = ", ".join([f"c.{v}" for v in variables])

        # For higher geographic levels, we need to aggregate (SUM) the block data
        # The census parquet files store data at block level, so we GROUP BY
        # the truncated GEOID to get tract/county/state totals
        agg_vars = ", ".join([f"SUM({v}) as {v}" for v in variables])

        sql = f"""
        SELECT
            i.GEOID,
            {var_list}
        FROM input_geoids i
        LEFT JOIN (
            SELECT LEFT(GEOID, {geoid_length}) as GEOID, {agg_vars}
            FROM {parquet_glob}
            GROUP BY LEFT(GEOID, {geoid_length})
        ) c ON i.GEOID = c.GEOID
        """

        result = self.query(sql)
        self.conn.unregister("input_geoids")

        return result

    def get_variables_for_geoid(
        self,
        geoid: str,
        variables: List[str],
        geo_level: Optional[GeoLevel] = None,
    ) -> Dict[str, Optional[float]]:
        """
        Get census variables for a single GEOID.

        Args:
            geoid: Geographic identifier
            variables: Variables to retrieve
            geo_level: Geographic level (auto-detected from GEOID length if not provided)

        Returns:
            Dictionary of variable values
        """
        # Auto-detect geo_level from GEOID length if not provided
        if geo_level is None:
            geo_level = GeoLevel.from_geoid_length(len(geoid))

        result = self.join_census_data([geoid], variables, geo_level)

        if result.empty:
            return {v: None for v in variables}

        row = result.iloc[0]
        return {v: row.get(v) for v in variables}



"""Main CensusLookup class - the primary user interface."""

import asyncio
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import geopandas as gpd
import pandas as pd
from shapely.geometry import Point

from census_lookup.address.matcher import TIGERAddressMatcher
from census_lookup.address.parser import AddressParser, ParsedAddress
from census_lookup.census.acs import (
    ACS_VARIABLE_GROUPS,
    ACS_VARIABLES,
    get_acs_variables_for_group,
)
from census_lookup.census.variables import VARIABLES, get_variables_for_group
from census_lookup.core.geoid import GEOIDParser, GeoLevel
from census_lookup.core.spatial import SpatialIndex
from census_lookup.data.constants import normalize_state
from census_lookup.data.manager import DataManager


@dataclass
class LookupResult:
    """Result from a single address/coordinate lookup."""

    # Input
    input_address: Optional[str] = None
    parsed_address: Dict[str, Any] = field(default_factory=dict)

    # Geocoding results
    matched_address: Optional[str] = None
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    match_type: str = "no_match"  # "exact", "interpolated", "no_match"
    match_score: float = 0.0  # 0.0 to 1.0

    # Geographic identifiers
    geoid: Optional[str] = None  # Full GEOID at requested level
    state_fips: Optional[str] = None
    county_fips: Optional[str] = None
    tract: Optional[str] = None
    block_group: Optional[str] = None
    block: Optional[str] = None

    # Census data (dynamic based on selected variables)
    census_data: Dict[str, Any] = field(default_factory=dict)

    @property
    def is_matched(self) -> bool:
        """Check if the lookup was successful."""
        return self.geoid is not None

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "input_address": self.input_address,
            "matched_address": self.matched_address,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "match_type": self.match_type,
            "match_score": self.match_score,
            "geoid": self.geoid,
            "state_fips": self.state_fips,
            "county_fips": self.county_fips,
            "tract": self.tract,
            "block_group": self.block_group,
            "block": self.block,
            **self.census_data,
        }

    def to_series(self) -> pd.Series:
        """Convert to pandas Series."""
        return pd.Series(self.to_dict())


class CensusLookup:
    """
    Main interface for census-lookup library.

    All methods are async for efficient concurrent operation.

    Example usage:
        >>> lookup = CensusLookup()
        >>> await lookup.load_state("CA")  # Lazy download if needed
        >>> result = await lookup.geocode("123 Main St, Los Angeles, CA 90012")
        >>> print(result.geoid, result.census_data)
    """

    def __init__(
        self,
        data_dir: Optional[Path] = None,
        geo_level: GeoLevel = GeoLevel.BLOCK,
        variables: Optional[List[str]] = None,
        variable_groups: Optional[List[str]] = None,
        acs_variables: Optional[List[str]] = None,
        acs_variable_groups: Optional[List[str]] = None,
        auto_download: bool = True,
    ):
        """
        Initialize CensusLookup.

        Args:
            data_dir: Directory for cached data. Defaults to ~/.census-lookup/
            geo_level: Default geographic level for lookups
            variables: PL 94-171 Census variables (e.g., ["P1_001N", "P2_001N"])
            variable_groups: PL 94-171 variable groups (e.g., ["population", "housing"])
            acs_variables: ACS variables (e.g., ["B19013_001E", "B15003_022E"])
            acs_variable_groups: ACS variable groups (e.g., ["income", "education"])
            auto_download: Whether to automatically download missing data

        Note:
            ACS data is only available at tract level and above. If you request
            ACS variables with geo_level=BLOCK or BLOCK_GROUP, the ACS data will
            be joined at tract level.
        """
        self.geo_level = geo_level
        self.auto_download = auto_download

        # Initialize data manager
        self._data_manager = DataManager(
            data_dir=data_dir,
            auto_download=auto_download,
        )

        # Determine variables to use
        self._variables = self._resolve_variables(variables, variable_groups)
        self._acs_variables = self._resolve_acs_variables(acs_variables, acs_variable_groups)

        # State-specific components (loaded lazily)
        self._loaded_states: Dict[str, Dict] = {}

        # Shared components
        self._parser = AddressParser()

    async def close(self):
        """Close all async sessions."""
        await self._data_manager.close()

    def _resolve_variables(
        self,
        variables: Optional[List[str]],
        variable_groups: Optional[List[str]],
    ) -> List[str]:
        """Resolve PL 94-171 variable list from variables and/or groups."""
        result = set()

        if variables:
            result.update(variables)

        if variable_groups:
            for group in variable_groups:
                result.update(get_variables_for_group(group))

        if not result:
            # Default to basic population
            result.add("P1_001N")

        return sorted(result)

    def _resolve_acs_variables(
        self,
        acs_variables: Optional[List[str]],
        acs_variable_groups: Optional[List[str]],
    ) -> List[str]:
        """Resolve ACS variable list from variables and/or groups."""
        result = set()

        if acs_variables:
            result.update(acs_variables)

        if acs_variable_groups:
            for group in acs_variable_groups:
                result.update(get_acs_variables_for_group(group))

        return sorted(result)

    async def load_state(self, state: str, force_download: bool = False) -> None:
        """
        Load data for a state (downloads if not cached).

        Args:
            state: State name, abbreviation, or FIPS code
            force_download: Re-download even if cached
        """
        state_fips = normalize_state(state)

        if state_fips in self._loaded_states and not force_download:
            return

        # Ensure data is downloaded
        await self._data_manager.ensure_state_data(state_fips, show_progress=True)

        # Also download ACS data if ACS variables are requested
        if self._acs_variables:
            await self._data_manager.ensure_acs_data(
                state_fips,
                variables=self._acs_variables,
                geo_level="tract",  # ACS is at tract level
                show_progress=True,
            )

        # Load blocks for spatial lookup
        blocks = await self._data_manager.get_blocks(state_fips)
        spatial_index = SpatialIndex(blocks, geoid_column="GEOID20")

        # Load address features for geocoding
        addr_features = await self._data_manager.get_address_features(state_fips)
        geocoder = TIGERAddressMatcher(addr_features)

        self._loaded_states[state_fips] = {
            "spatial_index": spatial_index,
            "geocoder": geocoder,
        }

    async def load_states(self, states: List[str]) -> None:
        """Load multiple states concurrently."""
        await asyncio.gather(*[self.load_state(state) for state in states])

    async def _ensure_state_loaded(self, state_fips: str) -> None:
        """Ensure a state is loaded, loading it if necessary."""
        if state_fips not in self._loaded_states:
            if self.auto_download:
                await self.load_state(state_fips)
            else:
                raise ValueError(
                    f"State {state_fips} not loaded. "
                    "Call load_state() first or enable auto_download."
                )

    def _get_state_from_address(self, parsed: ParsedAddress) -> Optional[str]:
        """Extract state FIPS from parsed address."""
        if parsed.state:
            try:
                return normalize_state(parsed.state)
            except ValueError:
                pass
        return None

    async def geocode(
        self,
        address: str,
        geo_level: Optional[GeoLevel] = None,
    ) -> LookupResult:
        """
        Geocode a single address and return census data.

        Args:
            address: Full address string
            geo_level: Override default geographic level

        Returns:
            LookupResult with coordinates, GEOID, and census data
        """
        level = geo_level or self.geo_level

        # Parse address
        try:
            parsed = self._parser.parse(address)
        except Exception:
            return LookupResult(
                input_address=address,
                match_type="parse_error",
            )

        # Get state
        state_fips = self._get_state_from_address(parsed)
        if not state_fips:
            return LookupResult(
                input_address=address,
                parsed_address=parsed.to_dict(),
                match_type="no_state",
            )

        # Ensure state is loaded
        try:
            await self._ensure_state_loaded(state_fips)
        except ValueError:
            return LookupResult(
                input_address=address,
                parsed_address=parsed.to_dict(),
                match_type="state_not_loaded",
            )

        state_data = self._loaded_states[state_fips]

        # Geocode
        geocode_result = state_data["geocoder"].geocode_parsed(parsed)

        if not geocode_result.is_matched:
            return LookupResult(
                input_address=address,
                parsed_address=parsed.to_dict(),
                match_type="no_match",
            )

        # Spatial lookup
        point = Point(geocode_result.longitude, geocode_result.latitude)
        block_geoid = state_data["spatial_index"].lookup(point)

        if not block_geoid:
            return LookupResult(
                input_address=address,
                parsed_address=parsed.to_dict(),
                latitude=geocode_result.latitude,
                longitude=geocode_result.longitude,
                matched_address=geocode_result.matched_address,
                match_type="no_block",
            )

        # Truncate GEOID to requested level
        geoid = block_geoid[: level.geoid_length]

        # Parse GEOID components
        components = GEOIDParser.parse(block_geoid)

        # Get census data (PL 94-171)
        census_data = self._data_manager.duckdb.get_variables_for_geoid(
            geoid,
            self._variables,
        )

        # Get ACS data if requested (at tract level)
        if self._acs_variables:
            tract_geoid = block_geoid[:11]  # Truncate to tract level
            try:
                acs_df = await self._data_manager.get_acs_data(
                    components.state,
                    self._acs_variables,
                    GeoLevel.TRACT,
                )
                # Find matching tract
                acs_row = acs_df[acs_df["GEOID"] == tract_geoid]
                if not acs_row.empty:
                    for var in self._acs_variables:
                        if var in acs_row.columns:
                            value = acs_row[var].iloc[0]
                            census_data[var] = value if pd.notna(value) else None
            except Exception:
                # ACS data not available - continue without it
                pass

        return LookupResult(
            input_address=address,
            parsed_address=parsed.to_dict(),
            matched_address=geocode_result.matched_address,
            latitude=geocode_result.latitude,
            longitude=geocode_result.longitude,
            match_type=geocode_result.match_type,
            match_score=geocode_result.match_score,
            geoid=geoid,
            state_fips=components.state,
            county_fips=components.county_fips,
            tract=components.tract_geoid,
            block_group=components.block_group_geoid,
            block=block_geoid if level == GeoLevel.BLOCK else None,
            census_data=census_data,
        )

    async def geocode_batch(
        self,
        addresses: Union[List[str], pd.Series],
        geo_level: Optional[GeoLevel] = None,
        progress: bool = True,
    ) -> pd.DataFrame:
        """
        Geocode multiple addresses concurrently.

        Args:
            addresses: List or Series of address strings
            geo_level: Geographic level for results
            progress: Show progress bar

        Returns:
            DataFrame with original addresses and census data
        """
        if isinstance(addresses, pd.Series):
            addresses = addresses.tolist()

        # Geocode all addresses concurrently
        tasks = [self.geocode(address, geo_level) for address in addresses]

        if progress:
            from tqdm import tqdm

            results = []
            with tqdm(total=len(tasks), desc="Geocoding") as pbar:
                for coro in asyncio.as_completed(tasks):
                    result = await coro
                    results.append(result.to_dict())
                    pbar.update(1)
        else:
            results_list = await asyncio.gather(*tasks)
            results = [r.to_dict() for r in results_list]

        return pd.DataFrame(results)

    async def lookup_coordinates(
        self,
        lat: float,
        lon: float,
        geo_level: Optional[GeoLevel] = None,
    ) -> LookupResult:
        """
        Look up census data for a coordinate pair.

        Args:
            lat: Latitude (decimal degrees)
            lon: Longitude (decimal degrees)
            geo_level: Override default geographic level

        Returns:
            LookupResult with GEOID and census data
        """
        level = geo_level or self.geo_level

        # Determine state from coordinates (rough bounding box check)
        # For now, require state to be loaded first
        point = Point(lon, lat)

        # Try each loaded state
        block_geoid = None
        for state_fips, state_data in self._loaded_states.items():
            geoid = state_data["spatial_index"].lookup(point)
            if geoid:
                block_geoid = geoid
                break

        if not block_geoid:
            return LookupResult(
                latitude=lat,
                longitude=lon,
                match_type="no_block",
            )

        # Truncate to requested level
        geoid = block_geoid[: level.geoid_length]
        components = GEOIDParser.parse(block_geoid)

        # Get census data (PL 94-171)
        census_data = self._data_manager.duckdb.get_variables_for_geoid(
            geoid,
            self._variables,
        )

        # Get ACS data if requested (at tract level)
        if self._acs_variables:
            tract_geoid = block_geoid[:11]  # Truncate to tract level
            try:
                acs_df = await self._data_manager.get_acs_data(
                    components.state,
                    self._acs_variables,
                    GeoLevel.TRACT,
                )
                # Find matching tract
                acs_row = acs_df[acs_df["GEOID"] == tract_geoid]
                if not acs_row.empty:
                    for var in self._acs_variables:
                        if var in acs_row.columns:
                            value = acs_row[var].iloc[0]
                            census_data[var] = value if pd.notna(value) else None
            except Exception:
                # ACS data not available - continue without it
                pass

        return LookupResult(
            latitude=lat,
            longitude=lon,
            match_type="coordinates",
            match_score=1.0,
            geoid=geoid,
            state_fips=components.state,
            county_fips=components.county_fips,
            tract=components.tract_geoid,
            block_group=components.block_group_geoid,
            block=block_geoid if level == GeoLevel.BLOCK else None,
            census_data=census_data,
        )

    async def lookup_coordinates_batch(
        self,
        df: pd.DataFrame,
        lat_column: str = "latitude",
        lon_column: str = "longitude",
        geo_level: Optional[GeoLevel] = None,
    ) -> pd.DataFrame:
        """
        Batch lookup for coordinates in a DataFrame.

        Args:
            df: DataFrame with lat/lon columns
            lat_column: Name of latitude column
            lon_column: Name of longitude column
            geo_level: Geographic level for results

        Returns:
            DataFrame with GEOID and census data columns added
        """
        level = geo_level or self.geo_level

        # Create GeoSeries of points
        points = gpd.GeoSeries(
            [
                Point(lon, lat) if pd.notna(lat) and pd.notna(lon) else None
                for lat, lon in zip(df[lat_column], df[lon_column])
            ],
            crs="EPSG:4269",  # NAD 83
        )

        # Spatial join with each loaded state
        result = df.copy()
        result["_geoid"] = None

        for state_fips, state_data in self._loaded_states.items():
            geoids = state_data["spatial_index"].lookup_batch(points)
            # Fill in where we found matches
            mask = geoids["GEOID"].notna() & result["_geoid"].isna()
            result.loc[mask, "_geoid"] = geoids.loc[mask, "GEOID"]

        # Truncate to level
        result["GEOID"] = result["_geoid"].apply(
            lambda x: x[: level.geoid_length] if x else None
        )
        result = result.drop(columns=["_geoid"])

        # Join census data
        geoids = result["GEOID"].dropna().unique().tolist()
        if geoids:
            census_df = self._data_manager.duckdb.join_census_data(
                geoids,
                self._variables,
                level,
            )
            result = result.merge(census_df, on="GEOID", how="left")

        return result

    @property
    def loaded_states(self) -> List[str]:
        """List of currently loaded states."""
        return list(self._loaded_states.keys())

    @property
    def variables(self) -> List[str]:
        """List of selected census variables."""
        return self._variables

    @property
    def available_variables(self) -> Dict[str, str]:
        """Dictionary of all available census variables and descriptions."""
        return VARIABLES.copy()

    def set_variables(self, variables: List[str]) -> None:
        """Update the list of census variables to retrieve."""
        self._variables = sorted(set(variables))

    def add_variable_group(self, group: str) -> None:
        """Add PL 94-171 variables from a group to the current selection."""
        new_vars = get_variables_for_group(group)
        self._variables = sorted(set(self._variables) | set(new_vars))

    # ==========================================================================
    # ACS (American Community Survey) Support
    # ==========================================================================

    @property
    def acs_variables(self) -> List[str]:
        """List of selected ACS variables."""
        return self._acs_variables

    @property
    def available_acs_variables(self) -> Dict[str, str]:
        """Dictionary of all available ACS variables and descriptions."""
        return ACS_VARIABLES.copy()

    @property
    def available_acs_variable_groups(self) -> Dict[str, List[str]]:
        """Dictionary of ACS variable groups."""
        return ACS_VARIABLE_GROUPS.copy()

    def set_acs_variables(self, variables: List[str]) -> None:
        """Update the list of ACS variables to retrieve."""
        self._acs_variables = sorted(set(variables))

    def add_acs_variable_group(self, group: str) -> None:
        """Add ACS variables from a group to the current selection."""
        new_vars = get_acs_variables_for_group(group)
        self._acs_variables = sorted(set(self._acs_variables) | set(new_vars))

    def clear_acs_variables(self) -> None:
        """Clear all ACS variables."""
        self._acs_variables = []



"""Data management for census-lookup."""

from census_lookup.data.catalog import DataCatalog, DatasetInfo
from census_lookup.data.constants import FIPS_STATES, TIGER_URLS
from census_lookup.data.converter import GeoParquetConverter
from census_lookup.data.downloader import CensusDataDownloader, TIGERDownloader
from census_lookup.data.duckdb_engine import DuckDBEngine
from census_lookup.data.manager import DataManager

__all__ = [
    "DataManager",
    "DataCatalog",
    "DatasetInfo",
    "TIGERDownloader",
    "CensusDataDownloader",
    "GeoParquetConverter",
    "DuckDBEngine",
    "FIPS_STATES",
    "TIGER_URLS",
]



"""Core functionality for census-lookup."""

from census_lookup.core.geoid import GEOIDParser, GeoLevel
from census_lookup.core.lookup import CensusLookup, LookupResult

__all__ = [
    "CensusLookup",
    "LookupResult",
    "GeoLevel",
    "GEOIDParser",
]



"""Address parsing using usaddress library."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import usaddress


@dataclass
class ParsedAddress:
    """Parsed address components."""

    house_number: Optional[str] = None
    street_name_pre_directional: Optional[str] = None
    street_name_pre_modifier: Optional[str] = None
    street_name_pre_type: Optional[str] = None
    street_name: Optional[str] = None
    street_name_post_type: Optional[str] = None  # Ave, St, Blvd
    street_name_post_directional: Optional[str] = None
    subaddress_type: Optional[str] = None  # Apt, Suite, Unit
    subaddress_identifier: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    zipcode: Optional[str] = None

    # Raw components from usaddress
    raw_components: Dict[str, str] = field(default_factory=dict)

    @property
    def full_street_name(self) -> str:
        """Combine street name components into full street name."""
        parts = [
            self.street_name_pre_directional,
            self.street_name_pre_modifier,
            self.street_name_pre_type,
            self.street_name,
            self.street_name_post_type,
            self.street_name_post_directional,
        ]
        return " ".join(p for p in parts if p)

    @property
    def has_street_info(self) -> bool:
        """Check if we have enough street information for geocoding."""
        return bool(self.street_name and self.house_number)

    def to_dict(self) -> Dict[str, Optional[str]]:
        """Convert to dictionary."""
        return {
            "house_number": self.house_number,
            "street_name_pre_directional": self.street_name_pre_directional,
            "street_name": self.street_name,
            "street_name_post_type": self.street_name_post_type,
            "street_name_post_directional": self.street_name_post_directional,
            "subaddress_type": self.subaddress_type,
            "subaddress_identifier": self.subaddress_identifier,
            "city": self.city,
            "state": self.state,
            "zipcode": self.zipcode,
        }


class AddressParseError(Exception):
    """Error parsing an address."""

    def __init__(self, address: str, reason: str):
        self.address = address
        self.reason = reason
        super().__init__(f"Failed to parse '{address}': {reason}")


class AddressParser:
    """
    Parse unstructured address strings using usaddress library.

    usaddress uses CRF (Conditional Random Fields) for probabilistic
    parsing that handles edge cases better than rule-based parsers.
    """

    # Mapping from usaddress labels to our normalized labels
    LABEL_MAP = {
        "AddressNumber": "house_number",
        "AddressNumberPrefix": "house_number_prefix",
        "AddressNumberSuffix": "house_number_suffix",
        "StreetNamePreDirectional": "street_name_pre_directional",
        "StreetNamePreModifier": "street_name_pre_modifier",
        "StreetNamePreType": "street_name_pre_type",
        "StreetName": "street_name",
        "StreetNamePostType": "street_name_post_type",
        "StreetNamePostDirectional": "street_name_post_directional",
        "SubaddressType": "subaddress_type",
        "SubaddressIdentifier": "subaddress_identifier",
        "OccupancyType": "subaddress_type",
        "OccupancyIdentifier": "subaddress_identifier",
        "PlaceName": "city",
        "StateName": "state",
        "ZipCode": "zipcode",
    }

    def parse(self, address: str) -> ParsedAddress:
        """
        Parse an address string.

        Args:
            address: Full address string

        Returns:
            ParsedAddress with extracted components

        Raises:
            AddressParseError: If parsing fails completely
        """
        if not address or not address.strip():
            raise AddressParseError(address, "Empty address")

        try:
            tagged, _ = usaddress.tag(address)
            return self._to_parsed_address(tagged)
        except usaddress.RepeatedLabelError:
            # Handle repeated labels by using parse() instead
            parsed = usaddress.parse(address)
            return self._from_parse_result(parsed)
        except Exception as e:
            raise AddressParseError(address, str(e))

    def _to_parsed_address(self, tagged: Dict[str, str]) -> ParsedAddress:
        """Convert usaddress tagged dict to ParsedAddress."""
        result = ParsedAddress(raw_components=dict(tagged))

        for usaddress_label, value in tagged.items():
            our_label = self.LABEL_MAP.get(usaddress_label)
            if our_label and hasattr(result, our_label):
                setattr(result, our_label, value)

        return result

    def _from_parse_result(self, parsed: List[Tuple[str, str]]) -> ParsedAddress:
        """Convert usaddress parse() result (list of tuples) to ParsedAddress."""
        # Group by label, taking first occurrence
        grouped: Dict[str, str] = {}
        for value, label in parsed:
            if label not in grouped:
                grouped[label] = value
            elif label == "StreetName":
                # Concatenate street name parts
                grouped[label] += " " + value

        return self._to_parsed_address(grouped)



"""Download TIGER/Line and Census data files."""

import asyncio
import zipfile
from pathlib import Path
from typing import Callable, Dict, List, Optional

import aiohttp
from tqdm import tqdm

from census_lookup.data.constants import TIGER_URLS


class DownloadError(Exception):
    """Error downloading a file."""

    def __init__(self, url: str, status_code: int, message: str = ""):
        self.url = url
        self.status_code = status_code
        super().__init__(f"Failed to download {url}: HTTP {status_code}. {message}")


class DownloadCoordinator:
    """
    Coordinates concurrent downloads to prevent duplicate requests.

    Uses asyncio Events to allow multiple waiters to share a single download.
    When a download starts, other requesters wait for it to complete rather
    than starting duplicate downloads.
    """

    def __init__(self):
        self._lock = asyncio.Lock()
        self._pending: Dict[str, asyncio.Task] = {}

    async def download_once(
        self,
        resource_key: str,
        download_func: Callable[[], "asyncio.Future[Path]"],
    ) -> Path:
        """
        Ensure a resource is downloaded only once, even with concurrent requests.

        If a download is already in progress for the resource, wait for it.
        Otherwise, start the download.

        Args:
            resource_key: Unique identifier for the resource (e.g., "blocks/06")
            download_func: Async function that performs the download

        Returns:
            Path to the downloaded resource
        """
        async with self._lock:
            # Check if there's already a pending download for this resource
            if resource_key in self._pending:
                task = self._pending[resource_key]
            else:
                # Start a new download task
                task = asyncio.create_task(download_func())
                self._pending[resource_key] = task

        try:
            # Wait for the task (either we created it or we're waiting on existing one)
            result = await task
            return result
        finally:
            # Clean up only if we're the one who started it and it's done
            async with self._lock:
                if resource_key in self._pending and self._pending[resource_key].done():
                    del self._pending[resource_key]


# Global coordinator instance
_coordinator = DownloadCoordinator()


def _extract_zip(zip_path: Path, extract_dir: Path) -> None:
    """Extract a zip file and clean up."""
    extract_dir.mkdir(parents=True, exist_ok=True)
    with zipfile.ZipFile(zip_path, "r") as zf:
        zf.extractall(extract_dir)
    zip_path.unlink()


class TIGERDownloader:
    """
    Downloads TIGER/Line shapefiles from Census Bureau.

    Data sources:
    - Address Features: https://www2.census.gov/geo/tiger/TIGER2020/ADDRFEAT/
    - Blocks: https://www2.census.gov/geo/tiger/TIGER2020/TABBLOCK20/
    - Block Groups: https://www2.census.gov/geo/tiger/TIGER2020/BG/
    - Tracts: https://www2.census.gov/geo/tiger/TIGER2020/TRACT/

    All methods are async and use aiohttp for concurrent downloads.
    """

    def __init__(self, timeout: int = 300, retries: int = 3):
        """
        Initialize downloader.

        Args:
            timeout: Request timeout in seconds
            retries: Number of retry attempts for failed downloads
        """
        self.timeout = timeout
        self.retries = retries
        self._session: Optional[aiohttp.ClientSession] = None

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create aiohttp session."""
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(total=self.timeout)
            self._session = aiohttp.ClientSession(
                timeout=timeout,
                headers={"User-Agent": "census-lookup/0.1.0 (https://github.com/census-lookup)"},
            )
        return self._session

    async def close(self):
        """Close the aiohttp session."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def download_blocks(
        self,
        state_fips: str,
        dest_dir: Path,
    ) -> Path:
        """
        Download block shapefile for a state.

        Uses coordination to prevent duplicate downloads.

        Args:
            state_fips: 2-digit state FIPS code
            dest_dir: Destination directory

        Returns:
            Path to downloaded and extracted shapefile directory
        """
        url = TIGER_URLS["blocks"].format(state_fips=state_fips)
        resource_key = f"blocks/{state_fips}"

        async def do_download():
            return await self._download_and_extract(url, dest_dir)

        return await _coordinator.download_once(resource_key, do_download)

    async def download_address_features(
        self,
        county_fips: str,
        dest_dir: Path,
    ) -> Path:
        """
        Download address feature file for a county.

        Uses coordination to prevent duplicate downloads.

        Args:
            county_fips: 5-digit county FIPS code (state + county)
            dest_dir: Destination directory

        Returns:
            Path to downloaded and extracted shapefile directory
        """
        url = TIGER_URLS["addrfeat"].format(county_fips=county_fips)
        resource_key = f"addrfeat/{county_fips}"

        async def do_download():
            return await self._download_and_extract(url, dest_dir)

        return await _coordinator.download_once(resource_key, do_download)

    async def download_address_features_for_state(
        self,
        state_fips: str,
        county_fips_list: List[str],
        dest_dir: Path,
        max_concurrent: int = 10,
    ) -> List[Path]:
        """
        Download address feature files for all counties in a state.

        Downloads multiple counties concurrently for better performance.
        Uses coordination to prevent duplicate downloads.

        Args:
            state_fips: 2-digit state FIPS code
            county_fips_list: List of 5-digit county FIPS codes
            dest_dir: Destination directory
            max_concurrent: Maximum concurrent downloads

        Returns:
            List of paths to downloaded files
        """
        dest_dir.mkdir(parents=True, exist_ok=True)

        # Use semaphore to limit concurrent downloads
        semaphore = asyncio.Semaphore(max_concurrent)

        async def download_county(county_fips: str) -> Optional[Path]:
            async with semaphore:
                try:
                    return await self.download_address_features(county_fips, dest_dir)
                except DownloadError as e:
                    if e.status_code == 404:
                        return None
                    raise

        # Create tasks for all counties
        tasks = [download_county(fips) for fips in county_fips_list]

        # Execute with progress bar
        paths = []
        desc = f"Downloading address features for {state_fips}"
        with tqdm(total=len(tasks), desc=desc) as pbar:
            for coro in asyncio.as_completed(tasks):
                result = await coro
                if result is not None:
                    paths.append(result)
                pbar.update(1)

        return paths

    async def _download_and_extract(
        self,
        url: str,
        dest_dir: Path,
    ) -> Path:
        """
        Download a ZIP file and extract it.

        Args:
            url: URL to download
            dest_dir: Destination directory

        Returns:
            Path to extracted directory
        """
        from urllib.parse import urlparse

        dest_dir.mkdir(parents=True, exist_ok=True)

        # Determine output filename from URL
        filename = Path(urlparse(url).path).name
        zip_path = dest_dir / filename
        extract_dir = dest_dir / filename.replace(".zip", "")

        # Skip if already extracted
        if extract_dir.exists() and any(extract_dir.glob("*.shp")):
            return extract_dir

        # Download with retries
        for attempt in range(self.retries):
            try:
                await self._download_file(url, zip_path)
                break
            except aiohttp.ClientError as e:
                if attempt == self.retries - 1:
                    raise DownloadError(url, 0, str(e))

        # Extract (run in executor to not block event loop)
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, _extract_zip, zip_path, extract_dir)

        return extract_dir

    async def _download_file(
        self,
        url: str,
        dest_path: Path,
        chunk_size: int = 8192,
    ) -> Path:
        """
        Download a single file.

        Args:
            url: URL to download
            dest_path: Destination path
            chunk_size: Download chunk size

        Returns:
            Path to downloaded file
        """
        session = await self._get_session()

        async with session.get(url) as response:
            if response.status == 404:
                raise DownloadError(url, 404, "File not found")
            response.raise_for_status()

            with open(dest_path, "wb") as f:
                async for chunk in response.content.iter_chunked(chunk_size):
                    if chunk:
                        f.write(chunk)

        return dest_path


class CensusDataDownloader:
    """
    Downloads Census PL 94-171 data via the Census API.

    All methods are async and use aiohttp for concurrent downloads.
    Variable batches are downloaded concurrently when there are more
    than 50 variables (Census API limit per request).
    """

    API_BASE = "https://api.census.gov/data/2020/dec/pl"

    def __init__(self, timeout: int = 60):
        """
        Initialize Census data downloader.

        Args:
            timeout: Request timeout in seconds
        """
        self.timeout = timeout
        self._session: Optional[aiohttp.ClientSession] = None

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create aiohttp session."""
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(total=self.timeout)
            self._session = aiohttp.ClientSession(
                timeout=timeout,
                headers={"User-Agent": "census-lookup/0.1.0 (https://github.com/census-lookup)"},
            )
        return self._session

    async def close(self):
        """Close the aiohttp session."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def download_pl94171_for_state(
        self,
        state_fips: str,
        variables: List[str],
        geo_level: str,
        dest_path: Path,
    ) -> Path:
        """
        Download PL 94-171 data for a state via Census API.

        Uses coordination to prevent duplicate downloads.
        Variable batches are downloaded concurrently.

        Args:
            state_fips: 2-digit state FIPS code
            variables: List of variable codes to download
            geo_level: Geographic level (block, block group, tract, county)
            dest_path: Output path for CSV file

        Returns:
            Path to downloaded CSV file
        """
        resource_key = f"pl94171/{state_fips}/{geo_level}"

        async def do_download():
            return await self._download_pl94171(state_fips, variables, geo_level, dest_path)

        return await _coordinator.download_once(resource_key, do_download)

    async def _download_pl94171(
        self,
        state_fips: str,
        variables: List[str],
        geo_level: str,
        dest_path: Path,
    ) -> Path:
        """Internal implementation for PL 94-171 download."""
        import pandas as pd

        session = await self._get_session()
        geo_params = self._build_geo_params(state_fips, geo_level)

        # Batch variables into groups of 50
        var_batches = [variables[i : i + 50] for i in range(0, len(variables), 50)]

        # Download batches concurrently
        async def fetch_batch(batch: List[str]):
            params = {
                "get": ",".join(["GEO_ID"] + batch),
                "for": geo_params["for"],
                "in": geo_params.get("in", ""),
            }

            url = f"{self.API_BASE}?{'&'.join(f'{k}={v}' for k, v in params.items())}"
            async with session.get(url) as response:
                response.raise_for_status()
                return await response.json()

        tasks = [fetch_batch(batch) for batch in var_batches]
        all_data = await asyncio.gather(*tasks)

        # Merge batches
        dfs = []
        for data in all_data:
            df = pd.DataFrame(data[1:], columns=data[0])
            dfs.append(df)

        if len(dfs) > 1:
            result = dfs[0]
            for df in dfs[1:]:
                result = result.merge(df, on="GEO_ID", how="outer")
        else:
            result = dfs[0]

        result.to_csv(dest_path, index=False)
        return dest_path

    def _build_geo_params(self, state_fips: str, geo_level: str) -> dict:
        """Build geographic parameters for Census API.

        Note: PL 94-171 data is always downloaded at block level.
        Higher geographic levels are derived by aggregating block data.
        """
        if geo_level != "block":
            raise ValueError(
                f"PL 94-171 download only supports block level, got: {geo_level}. "
                "Higher levels are derived from block data."
            )

        return {
            "for": "block:*",
            "in": f"state:{state_fips} county:* tract:*",
        }


class ACSDataDownloader:
    """
    Downloads American Community Survey (ACS) 5-Year Estimates data.

    ACS provides richer demographic data than PL 94-171, including:
    - Income and poverty
    - Educational attainment
    - Employment and occupation
    - Housing characteristics
    - Health insurance
    - Commute patterns

    Note: ACS data is available at tract level and above, not at block level.
    For block-level data, use PL 94-171.

    All methods are async and use aiohttp for concurrent downloads.
    """

    def __init__(
        self,
        timeout: int = 120,
        year: int = 2020,
    ):
        """
        Initialize ACS data downloader.

        Args:
            timeout: Request timeout in seconds
            year: ACS year (default 2020)
        """
        self.timeout = timeout
        self.year = year
        self._session: Optional[aiohttp.ClientSession] = None

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create aiohttp session."""
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(total=self.timeout)
            self._session = aiohttp.ClientSession(
                timeout=timeout,
                headers={"User-Agent": "census-lookup/0.1.0 (https://github.com/census-lookup)"},
            )
        return self._session

    async def close(self):
        """Close the aiohttp session."""
        if self._session and not self._session.closed:
            await self._session.close()

    @property
    def api_base(self) -> str:
        """Get the API base URL for the configured year."""
        return f"https://api.census.gov/data/{self.year}/acs/acs5"

    async def download_acs_for_state(
        self,
        state_fips: str,
        variables: List[str],
        geo_level: str,
        dest_path: Path,
    ) -> Path:
        """
        Download ACS 5-Year data for a state via Census API.

        Uses coordination to prevent duplicate downloads.
        Variable batches are downloaded concurrently.

        Args:
            state_fips: 2-digit state FIPS code
            variables: List of variable codes to download
            geo_level: Geographic level (tract, block group, county)
                       Note: ACS is not available at block level
            dest_path: Output path for CSV file

        Returns:
            Path to downloaded CSV file
        """
        resource_key = f"acs/{state_fips}/{geo_level}/{self.year}"

        async def do_download():
            return await self._download_acs(state_fips, variables, geo_level, dest_path)

        return await _coordinator.download_once(resource_key, do_download)

    async def _download_acs(
        self,
        state_fips: str,
        variables: List[str],
        geo_level: str,
        dest_path: Path,
    ) -> Path:
        """Internal implementation for ACS download."""
        import pandas as pd

        # Validate geo_level
        valid_levels = ["tract", "block group", "county"]
        if geo_level not in valid_levels:
            raise ValueError(
                f"Invalid geo_level for ACS: {geo_level}. "
                f"ACS is available at: {', '.join(valid_levels)}"
            )

        session = await self._get_session()
        geo_params = self._build_geo_params(state_fips, geo_level)

        # Batch variables into groups of 50 (Census API limit)
        var_batches = [variables[i : i + 50] for i in range(0, len(variables), 50)]

        async def fetch_batch(batch: List[str]):
            params = {
                "get": ",".join(["GEO_ID", "NAME"] + batch),
                "for": geo_params["for"],
                "in": geo_params.get("in", ""),
            }

            url = f"{self.api_base}?{'&'.join(f'{k}={v}' for k, v in params.items())}"
            async with session.get(url) as response:
                if response.status == 400:
                    error_msg = await response.text()
                    raise DownloadError(
                        self.api_base,
                        400,
                        f"Invalid API request. Check variable names. {error_msg}",
                    )
                response.raise_for_status()
                return await response.json()

        # Fetch all batches concurrently with progress
        all_data = []
        desc = f"Downloading ACS data for {state_fips}"
        with tqdm(total=len(var_batches), desc=desc) as pbar:
            for coro in asyncio.as_completed([fetch_batch(b) for b in var_batches]):
                result = await coro
                all_data.append(result)
                pbar.update(1)

        # Merge batches
        dfs = []
        for data in all_data:
            df = pd.DataFrame(data[1:], columns=data[0])
            dfs.append(df)

        if len(dfs) > 1:
            result = dfs[0]
            for df in dfs[1:]:
                df_no_name = df.drop(columns=["NAME"], errors="ignore")
                result = result.merge(df_no_name, on="GEO_ID", how="outer")
        else:
            result = dfs[0]

        # Clean up GEOID format
        result["GEOID"] = result["GEO_ID"].str.extract(r"(\d+)$")[0]

        # Convert numeric columns
        for col in variables:
            if col in result.columns:
                result[col] = pd.to_numeric(result[col], errors="coerce")

        dest_path.parent.mkdir(parents=True, exist_ok=True)
        result.to_csv(dest_path, index=False)
        return dest_path

    def _build_geo_params(self, state_fips: str, geo_level: str) -> dict:
        """Build geographic parameters for Census API."""
        params = {}

        if geo_level == "block group":
            params["for"] = "block group:*"
            params["in"] = f"state:{state_fips} county:* tract:*"
        elif geo_level == "tract":
            params["for"] = "tract:*"
            params["in"] = f"state:{state_fips} county:*"
        elif geo_level == "county":
            params["for"] = "county:*"
            params["in"] = f"state:{state_fips}"
        else:
            raise ValueError(f"Unknown geo_level: {geo_level}")

        return params



"""
census-lookup: Offline address-to-Census 2020 data mapping for Python.

Map US addresses to Census 2020 block-level data locally, avoiding API rate limits.

Supports:
- PL 94-171 (Redistricting Data): Population, race, housing
- ACS 5-Year Estimates: Income, education, employment, housing characteristics
"""

from census_lookup.census.acs import (
    ACS_VARIABLE_GROUPS,
    ACS_VARIABLES,
    get_acs_variables_for_group,
    list_acs_tables,
    list_acs_variable_groups,
)
from census_lookup.census.variables import (
    VARIABLE_GROUPS,
    VARIABLES,
    get_variables_for_group,
    list_tables,
    list_variable_groups,
)
from census_lookup.core.geoid import GeoLevel
from census_lookup.core.lookup import CensusLookup, LookupResult
from census_lookup.data.downloader import DownloadError

__version__ = "0.1.0"
__all__ = [
    # Main API
    "CensusLookup",
    "LookupResult",
    "GeoLevel",
    # Exceptions
    "DownloadError",
    # PL 94-171 Variables
    "VARIABLES",
    "VARIABLE_GROUPS",
    "get_variables_for_group",
    "list_tables",
    "list_variable_groups",
    # ACS Variables
    "ACS_VARIABLES",
    "ACS_VARIABLE_GROUPS",
    "get_acs_variables_for_group",
    "list_acs_tables",
    "list_acs_variable_groups",
]



"""Spatial index and point-in-polygon operations."""

from typing import Optional

import geopandas as gpd
import pandas as pd
from shapely.geometry import Point


class SpatialIndex:
    """
    Efficient spatial index for point-in-polygon lookups.

    Uses GeoPandas sindex (backed by rtree) for ~17x speedup
    over naive iteration.
    """

    def __init__(
        self,
        polygons: gpd.GeoDataFrame,
        geoid_column: str = "GEOID20",
    ):
        """
        Initialize spatial index from polygon GeoDataFrame.

        Args:
            polygons: GeoDataFrame with polygon geometries
            geoid_column: Column name containing GEOID
        """
        self._polygons = polygons
        self._geoid_col = geoid_column

        # Ensure spatial index is built
        if not self._polygons.has_sindex:
            _ = self._polygons.sindex

    def lookup(self, point: Point) -> Optional[str]:
        """
        Find GEOID containing a point.

        Uses two-phase approach:
        1. Query spatial index for candidate polygons (bbox intersection)
        2. Exact point-in-polygon test on candidates

        Args:
            point: Shapely Point object

        Returns:
            GEOID string if found, None otherwise
        """
        # Get candidate indices from spatial index
        candidates = list(self._polygons.sindex.query(point, predicate="intersects"))

        if len(candidates) == 0:
            return None

        # For census blocks, should typically be exactly one match
        # Use covers() instead of contains() to include boundary points
        # If multiple match (point on boundary), return smallest by area
        matches = []
        for idx in candidates:
            row = self._polygons.iloc[idx]
            geom = row.geometry
            # covers() returns True if point is inside OR on boundary
            if geom.covers(point):
                matches.append((row[self._geoid_col], geom.area))

        if not matches:
            return None

        # If multiple matches (boundary point), prefer smaller area block
        # This typically gives more specific result
        matches.sort(key=lambda x: x[1])
        return matches[0][0]

    def lookup_coordinates(
        self,
        lat: float,
        lon: float,
    ) -> Optional[str]:
        """
        Look up GEOID for lat/lon coordinates.

        Args:
            lat: Latitude (decimal degrees)
            lon: Longitude (decimal degrees)

        Returns:
            GEOID string if found, None otherwise
        """
        point = Point(lon, lat)  # Note: Shapely uses (x, y) = (lon, lat)
        return self.lookup(point)

    def lookup_batch(
        self,
        points: gpd.GeoSeries,
    ) -> pd.DataFrame:
        """
        Batch lookup using spatial join.

        Much faster than iterating for large point sets.
        Uses gpd.sjoin with "within" predicate.

        Args:
            points: GeoSeries of Point geometries

        Returns:
            DataFrame with GEOID for each input point
        """
        # Create GeoDataFrame from points
        points_gdf = gpd.GeoDataFrame(
            {"_idx": range(len(points))},
            geometry=points,
            crs=self._polygons.crs,
        )

        # Spatial join
        result = gpd.sjoin(
            points_gdf,
            self._polygons[[self._geoid_col, "geometry"]],
            how="left",
            predicate="within",
        )

        # Handle duplicates (shouldn't happen with census blocks, but be safe)
        result = result.drop_duplicates(subset="_idx", keep="first")

        # Ensure we have all original points
        result = result.set_index("_idx").reindex(range(len(points)))

        return result[[self._geoid_col]].rename(columns={self._geoid_col: "GEOID"})

    @property
    def crs(self):
        """Get coordinate reference system."""
        return self._polygons.crs



"""Convert shapefiles to GeoParquet format."""

from pathlib import Path
from typing import List, Optional

import geopandas as gpd
import pandas as pd


class GeoParquetConverter:
    """
    Converts shapefiles to GeoParquet for efficient storage and loading.

    Benefits:
    - ~60% smaller file size
    - 2-4x faster reading
    - Column-selective reads
    - Spatial filtering support
    """

    def __init__(self, compression: str = "zstd"):
        """
        Initialize converter.

        Args:
            compression: Compression algorithm (zstd, snappy, gzip, or None)
        """
        self.compression = compression

    def convert_shapefile(
        self,
        shapefile_path: Path,
        output_path: Path,
        columns: Optional[List[str]] = None,
    ) -> Path:
        """
        Convert a shapefile to GeoParquet.

        Args:
            shapefile_path: Input shapefile path (.shp or directory containing .shp)
            output_path: Output parquet path
            columns: Columns to include (None for all)

        Returns:
            Path to output parquet file
        """
        # Handle both .shp file and directory containing .shp
        if shapefile_path.is_dir():
            shp_files = list(shapefile_path.glob("*.shp"))
            if not shp_files:
                raise FileNotFoundError(f"No .shp file found in {shapefile_path}")
            shapefile_path = shp_files[0]

        gdf = gpd.read_file(shapefile_path)

        if columns:
            # Always include geometry
            columns = list(set(columns) | {"geometry"})
            # Filter to existing columns
            existing_cols = [c for c in columns if c in gdf.columns]
            gdf = gdf[existing_cols]

        output_path.parent.mkdir(parents=True, exist_ok=True)
        gdf.to_parquet(output_path, compression=self.compression)

        return output_path

    def convert_blocks(
        self,
        shapefile_path: Path,
        output_path: Path,
    ) -> Path:
        """
        Convert block shapefile to GeoParquet with essential columns only.

        Essential columns for blocks:
        - GEOID20: Block GEOID (15-digit string)
        - STATEFP20: State FIPS
        - COUNTYFP20: County FIPS
        - TRACTCE20: Tract code
        - BLOCKCE20: Block code
        - geometry: Polygon geometry

        Raises:
            ValueError: If GEOID20 values are not valid 15-digit strings
        """
        columns = [
            "GEOID20",
            "STATEFP20",
            "COUNTYFP20",
            "TRACTCE20",
            "BLOCKCE20",
            "ALAND20",  # Land area
            "AWATER20",  # Water area
            "geometry",
        ]

        # Load and validate
        if shapefile_path.is_dir():
            shp_files = list(shapefile_path.glob("*.shp"))
            if not shp_files:
                raise FileNotFoundError(f"No .shp file found in {shapefile_path}")
            shapefile_path = shp_files[0]

        gdf = gpd.read_file(shapefile_path)

        # Validate GEOID20 format
        invalid_geoids = gdf[
            ~gdf["GEOID20"].str.match(r"^\d{15}$", na=False)
        ]
        if not invalid_geoids.empty:
            samples = invalid_geoids["GEOID20"].head(5).tolist()
            raise ValueError(
                f"Invalid GEOID20 values in block data (expected 15 digits): {samples}"
            )

        # Filter columns
        columns = list(set(columns) | {"geometry"})
        existing_cols = [c for c in columns if c in gdf.columns]
        gdf = gdf[existing_cols]

        output_path.parent.mkdir(parents=True, exist_ok=True)
        gdf.to_parquet(output_path, compression=self.compression)

        return output_path

    def convert_address_features(
        self,
        shapefile_path: Path,
        output_path: Path,
    ) -> Path:
        """
        Convert address feature shapefile to GeoParquet.

        Essential columns:
        - FULLNAME: Full street name
        - LFROMHN, LTOHN: Left side house number range
        - RFROMHN, RTOHN: Right side house number range
        - ZIPL, ZIPR: ZIP codes for each side
        - LINEARID: Feature ID
        - geometry: LineString geometry
        """
        columns = [
            "LINEARID",
            "FULLNAME",
            "LFROMHN",
            "LTOHN",
            "RFROMHN",
            "RTOHN",
            "ZIPL",
            "ZIPR",
            "PARITYL",
            "PARITYR",
            "geometry",
        ]
        return self.convert_shapefile(shapefile_path, output_path, columns)

    def merge_county_files(
        self,
        county_files: List[Path],
        output_path: Path,
    ) -> Path:
        """
        Merge multiple county-level parquet files into a single state-level file.

        Args:
            county_files: List of county-level parquet files
            output_path: Output state-level parquet file

        Returns:
            Path to merged output file
        """
        if not county_files:
            raise ValueError("No county files provided")

        gdfs = []
        for path in county_files:
            if path.suffix == ".parquet":
                gdf = gpd.read_parquet(path)
            else:
                gdf = gpd.read_file(path)
            gdfs.append(gdf)

        merged = gpd.GeoDataFrame(pd.concat(gdfs, ignore_index=True))

        output_path.parent.mkdir(parents=True, exist_ok=True)
        merged.to_parquet(output_path, compression=self.compression)

        return output_path

    def convert_census_csv(
        self,
        csv_path: Path,
        output_path: Path,
        geoid_column: str = "GEO_ID",
    ) -> Path:
        """
        Convert Census CSV data to Parquet.

        Args:
            csv_path: Input CSV path
            output_path: Output parquet path
            geoid_column: Column containing GEOID

        Returns:
            Path to output parquet file
        """
        df = pd.read_csv(csv_path, dtype={geoid_column: str})

        # Clean up GEO_ID format (Census API returns "1000000US{GEOID}")
        if geoid_column in df.columns:
            df["GEOID"] = df[geoid_column].str.extract(r"(\d+)$")

        output_path.parent.mkdir(parents=True, exist_ok=True)
        df.to_parquet(output_path, compression=self.compression)

        return output_path



"""Street name normalization for TIGER matching."""

import re
from typing import Dict


class StreetNormalizer:
    """
    Normalize street names for matching against TIGER data.

    Handles:
    - Directional abbreviations (N -> NORTH)
    - Street type variations (ST -> STREET, AVE -> AVENUE)
    - Ordinal numbers (1ST -> FIRST)
    - Case normalization
    - Special characters
    """

    # Directional abbreviations to full form
    DIRECTIONALS: Dict[str, str] = {
        "N": "NORTH",
        "S": "SOUTH",
        "E": "EAST",
        "W": "WEST",
        "NE": "NORTHEAST",
        "NW": "NORTHWEST",
        "SE": "SOUTHEAST",
        "SW": "SOUTHWEST",
        "NO": "NORTH",
        "SO": "SOUTH",
    }

    # Reverse mapping: full form to abbreviation (for TIGER matching)
    DIRECTIONALS_ABBREV: Dict[str, str] = {
        "NORTH": "N",
        "SOUTH": "S",
        "EAST": "E",
        "WEST": "W",
        "NORTHEAST": "NE",
        "NORTHWEST": "NW",
        "SOUTHEAST": "SE",
        "SOUTHWEST": "SW",
    }

    # Street type abbreviations to full form
    STREET_TYPES: Dict[str, str] = {
        "ST": "STREET",
        "STR": "STREET",
        "AVE": "AVENUE",
        "AV": "AVENUE",
        "BLVD": "BOULEVARD",
        "BLV": "BOULEVARD",
        "DR": "DRIVE",
        "DRV": "DRIVE",
        "RD": "ROAD",
        "LN": "LANE",
        "CT": "COURT",
        "CRT": "COURT",
        "PL": "PLACE",
        "WAY": "WAY",
        "CIR": "CIRCLE",
        "CRCL": "CIRCLE",
        "TRL": "TRAIL",
        "TR": "TRAIL",
        "PKWY": "PARKWAY",
        "PKY": "PARKWAY",
        "HWY": "HIGHWAY",
        "HWAY": "HIGHWAY",
        "EXPY": "EXPRESSWAY",
        "EXP": "EXPRESSWAY",
        "EXPW": "EXPRESSWAY",
        "FWY": "FREEWAY",
        "FRWY": "FREEWAY",
        "ALY": "ALLEY",
        "ALLY": "ALLEY",
        "ANX": "ANNEX",
        "ARC": "ARCADE",
        "BCH": "BEACH",
        "BND": "BEND",
        "BRG": "BRIDGE",
        "BRK": "BROOK",
        "BYP": "BYPASS",
        "CYN": "CANYON",
        "CPE": "CAPE",
        "CSWY": "CAUSEWAY",
        "CTR": "CENTER",
        "CLF": "CLIFF",
        "CLB": "CLUB",
        "CMN": "COMMON",
        "CMNS": "COMMONS",
        "CRK": "CREEK",
        "CRES": "CRESCENT",
        "CRST": "CREST",
        "XING": "CROSSING",
        "DL": "DALE",
        "DM": "DAM",
        "DV": "DIVIDE",
        "EST": "ESTATE",
        "ESTS": "ESTATES",
        "FALL": "FALL",
        "FLS": "FALLS",
        "FRY": "FERRY",
        "FLD": "FIELD",
        "FLDS": "FIELDS",
        "FLT": "FLAT",
        "FLTS": "FLATS",
        "FRD": "FORD",
        "FRST": "FOREST",
        "FRG": "FORGE",
        "FRK": "FORK",
        "FRKS": "FORKS",
        "FT": "FORT",
        "GDN": "GARDEN",
        "GDNS": "GARDENS",
        "GTWY": "GATEWAY",
        "GLN": "GLEN",
        "GRN": "GREEN",
        "GRV": "GROVE",
        "HBR": "HARBOR",
        "HVN": "HAVEN",
        "HTS": "HEIGHTS",
        "HL": "HILL",
        "HLS": "HILLS",
        "HOLW": "HOLLOW",
        "INLT": "INLET",
        "IS": "ISLAND",
        "ISS": "ISLANDS",
        "JCT": "JUNCTION",
        "KY": "KEY",
        "KYS": "KEYS",
        "KNL": "KNOLL",
        "KNLS": "KNOLLS",
        "LK": "LAKE",
        "LKS": "LAKES",
        "LNDG": "LANDING",
        "LGT": "LIGHT",
        "LF": "LOAF",
        "LCK": "LOCK",
        "LCKS": "LOCKS",
        "LDG": "LODGE",
        "LOOP": "LOOP",
        "MALL": "MALL",
        "MNR": "MANOR",
        "MDWS": "MEADOWS",
        "ML": "MILL",
        "MLS": "MILLS",
        "MSN": "MISSION",
        "MT": "MOUNT",
        "MTN": "MOUNTAIN",
        "NCK": "NECK",
        "ORCH": "ORCHARD",
        "OVAL": "OVAL",
        "PARK": "PARK",
        "PASS": "PASS",
        "PATH": "PATH",
        "PIKE": "PIKE",
        "PNE": "PINE",
        "PNES": "PINES",
        "PLN": "PLAIN",
        "PLNS": "PLAINS",
        "PLZ": "PLAZA",
        "PT": "POINT",
        "PTS": "POINTS",
        "PRT": "PORT",
        "PRTS": "PORTS",
        "PR": "PRAIRIE",
        "RADL": "RADIAL",
        "RNCH": "RANCH",
        "RPD": "RAPID",
        "RPDS": "RAPIDS",
        "RST": "REST",
        "RDG": "RIDGE",
        "RDGS": "RIDGES",
        "RIV": "RIVER",
        "ROW": "ROW",
        "RUN": "RUN",
        "SHL": "SHOAL",
        "SHLS": "SHOALS",
        "SHR": "SHORE",
        "SHRS": "SHORES",
        "SPG": "SPRING",
        "SPGS": "SPRINGS",
        "SPUR": "SPUR",
        "SQ": "SQUARE",
        "SQS": "SQUARES",
        "STA": "STATION",
        "STRA": "STRAVENUE",
        "STRM": "STREAM",
        "SMT": "SUMMIT",
        "TER": "TERRACE",
        "TRCE": "TRACE",
        "TRAK": "TRACK",
        "TRFY": "TRAFFICWAY",
        "TUNL": "TUNNEL",
        "TPKE": "TURNPIKE",
        "UN": "UNION",
        "UNS": "UNIONS",
        "VLY": "VALLEY",
        "VLYS": "VALLEYS",
        "VIA": "VIADUCT",
        "VW": "VIEW",
        "VWS": "VIEWS",
        "VLG": "VILLAGE",
        "VLGS": "VILLAGES",
        "VL": "VILLE",
        "VIS": "VISTA",
        "WALK": "WALK",
        "WALL": "WALL",
        "WL": "WELL",
        "WLS": "WELLS",
    }

    # Reverse mapping: full form to preferred TIGER abbreviation
    # Note: TIGER data uses specific abbreviations, we select the most common ones
    STREET_TYPES_ABBREV: Dict[str, str] = {
        "STREET": "ST",
        "AVENUE": "AVE",
        "BOULEVARD": "BLVD",
        "DRIVE": "DR",
        "ROAD": "RD",
        "LANE": "LN",
        "COURT": "CT",
        "PLACE": "PL",
        "WAY": "WAY",
        "CIRCLE": "CIR",
        "TRAIL": "TRL",
        "PARKWAY": "PKWY",
        "HIGHWAY": "HWY",
        "EXPRESSWAY": "EXPY",
        "FREEWAY": "FWY",
        "ALLEY": "ALY",
        "ANNEX": "ANX",
        "ARCADE": "ARC",
        "BEACH": "BCH",
        "BEND": "BND",
        "BRIDGE": "BRG",
        "BROOK": "BRK",
        "BYPASS": "BYP",
        "CANYON": "CYN",
        "CAPE": "CPE",
        "CAUSEWAY": "CSWY",
        "CENTER": "CTR",
        "CLIFF": "CLF",
        "CLUB": "CLB",
        "COMMON": "CMN",
        "COMMONS": "CMNS",
        "CREEK": "CRK",
        "CRESCENT": "CRES",
        "CREST": "CRST",
        "CROSSING": "XING",
        "DALE": "DL",
        "DAM": "DM",
        "DIVIDE": "DV",
        "ESTATE": "EST",
        "ESTATES": "ESTS",
        "FALLS": "FLS",
        "FERRY": "FRY",
        "FIELD": "FLD",
        "FIELDS": "FLDS",
        "FLAT": "FLT",
        "FLATS": "FLTS",
        "FORD": "FRD",
        "FOREST": "FRST",
        "FORGE": "FRG",
        "FORK": "FRK",
        "FORKS": "FRKS",
        "FORT": "FT",
        "GARDEN": "GDN",
        "GARDENS": "GDNS",
        "GATEWAY": "GTWY",
        "GLEN": "GLN",
        "GREEN": "GRN",
        "GROVE": "GRV",
        "HARBOR": "HBR",
        "HAVEN": "HVN",
        "HEIGHTS": "HTS",
        "HILL": "HL",
        "HILLS": "HLS",
        "HOLLOW": "HOLW",
        "INLET": "INLT",
        "ISLAND": "IS",
        "ISLANDS": "ISS",
        "JUNCTION": "JCT",
        "KEY": "KY",
        "KEYS": "KYS",
        "KNOLL": "KNL",
        "KNOLLS": "KNLS",
        "LAKE": "LK",
        "LAKES": "LKS",
        "LANDING": "LNDG",
        "LIGHT": "LGT",
        "LOAF": "LF",
        "LOCK": "LCK",
        "LOCKS": "LCKS",
        "LODGE": "LDG",
        "LOOP": "LOOP",
        "MALL": "MALL",
        "MANOR": "MNR",
        "MEADOWS": "MDWS",
        "MILL": "ML",
        "MILLS": "MLS",
        "MISSION": "MSN",
        "MOUNT": "MT",
        "MOUNTAIN": "MTN",
        "NECK": "NCK",
        "ORCHARD": "ORCH",
        "OVAL": "OVAL",
        "PARK": "PARK",
        "PASS": "PASS",
        "PATH": "PATH",
        "PIKE": "PIKE",
        "PINE": "PNE",
        "PINES": "PNES",
        "PLAIN": "PLN",
        "PLAINS": "PLNS",
        "PLAZA": "PLZ",
        "POINT": "PT",
        "POINTS": "PTS",
        "PORT": "PRT",
        "PORTS": "PRTS",
        "PRAIRIE": "PR",
        "RADIAL": "RADL",
        "RANCH": "RNCH",
        "RAPID": "RPD",
        "RAPIDS": "RPDS",
        "REST": "RST",
        "RIDGE": "RDG",
        "RIDGES": "RDGS",
        "RIVER": "RIV",
        "ROW": "ROW",
        "RUN": "RUN",
        "SHOAL": "SHL",
        "SHOALS": "SHLS",
        "SHORE": "SHR",
        "SHORES": "SHRS",
        "SPRING": "SPG",
        "SPRINGS": "SPGS",
        "SPUR": "SPUR",
        "SQUARE": "SQ",
        "SQUARES": "SQS",
        "STATION": "STA",
        "STRAVENUE": "STRA",
        "STREAM": "STRM",
        "SUMMIT": "SMT",
        "TERRACE": "TER",
        "TRACE": "TRCE",
        "TRACK": "TRAK",
        "TRAFFICWAY": "TRFY",
        "TUNNEL": "TUNL",
        "TURNPIKE": "TPKE",
        "UNION": "UN",
        "UNIONS": "UNS",
        "VALLEY": "VLY",
        "VALLEYS": "VLYS",
        "VIADUCT": "VIA",
        "VIEW": "VW",
        "VIEWS": "VWS",
        "VILLAGE": "VLG",
        "VILLAGES": "VLGS",
        "VILLE": "VL",
        "VISTA": "VIS",
        "WALK": "WALK",
        "WALL": "WALL",
        "WELL": "WL",
        "WELLS": "WLS",
    }

    # Ordinal numbers
    ORDINALS: Dict[str, str] = {
        "1ST": "FIRST",
        "2ND": "SECOND",
        "3RD": "THIRD",
        "4TH": "FOURTH",
        "5TH": "FIFTH",
        "6TH": "SIXTH",
        "7TH": "SEVENTH",
        "8TH": "EIGHTH",
        "9TH": "NINTH",
        "10TH": "TENTH",
        "11TH": "ELEVENTH",
        "12TH": "TWELFTH",
    }

    def normalize(
        self,
        street_name: str,
        expand_abbreviations: bool = True,
    ) -> str:
        """
        Normalize a street name for matching.

        Args:
            street_name: Street name to normalize
            expand_abbreviations: Whether to expand abbreviations to full form

        Returns:
            Normalized uppercase street name
        """
        if not street_name:
            return ""

        # Uppercase
        result = street_name.upper().strip()

        # Remove extra whitespace
        result = " ".join(result.split())

        # Remove special characters except spaces and hyphens
        result = re.sub(r"[^\w\s\-]", "", result)

        if expand_abbreviations:
            # Expand directionals
            words = result.split()
            normalized_words = []
            for word in words:
                if word in self.DIRECTIONALS:
                    normalized_words.append(self.DIRECTIONALS[word])
                elif word in self.STREET_TYPES:
                    normalized_words.append(self.STREET_TYPES[word])
                elif word in self.ORDINALS:
                    normalized_words.append(self.ORDINALS[word])
                else:
                    normalized_words.append(word)
            result = " ".join(normalized_words)

        return result

    def generate_variants(self, street_name: str) -> list:
        """
        Generate possible variants of a street name for fuzzy matching.

        Args:
            street_name: Normalized street name

        Returns:
            List of variant strings to try matching
        """
        variants = [street_name]

        words = street_name.split()

        # Try converting full street types to abbreviations (AVENUE -> AVE)
        # This is important for matching TIGER data which uses abbreviations
        new_words = []
        changed = False
        for word in words:
            if word in self.STREET_TYPES_ABBREV:
                # For common types, prefer the first (canonical) abbreviation
                abbrev = self.STREET_TYPES_ABBREV[word]
                new_words.append(abbrev)
                changed = True
            else:
                new_words.append(word)
        if changed:
            variants.append(" ".join(new_words))

        # Try with abbreviated directionals (NORTHWEST -> NW)
        new_words = []
        changed = False
        for word in words:
            if word in self.DIRECTIONALS_ABBREV:
                new_words.append(self.DIRECTIONALS_ABBREV[word])
                changed = True
            else:
                new_words.append(word)
        if changed:
            variants.append(" ".join(new_words))

        # Try both street type AND directional abbreviations together
        new_words = []
        for word in words:
            if word in self.STREET_TYPES_ABBREV:
                new_words.append(self.STREET_TYPES_ABBREV[word])
            elif word in self.DIRECTIONALS_ABBREV:
                new_words.append(self.DIRECTIONALS_ABBREV[word])
            else:
                new_words.append(word)
        combined = " ".join(new_words)
        if combined not in variants:
            variants.append(combined)

        # Try without street type (last word if it's a type)
        if len(words) > 1:
            last_word = words[-1]
            if last_word in self.STREET_TYPES_ABBREV or last_word in self.STREET_TYPES:
                variants.append(" ".join(words[:-1]))

        return variants



"""Census data handling for census-lookup."""

from census_lookup.census.variables import (
    DEFAULT_VARIABLES,
    VARIABLE_GROUPS,
    VARIABLES,
    get_variables_for_group,
)

__all__ = [
    "VARIABLES",
    "VARIABLE_GROUPS",
    "DEFAULT_VARIABLES",
    "get_variables_for_group",
]
